"""CobraMod's helpful function

This module is intended to stored multiple functions, which cannot be grouped
to any module. In other words, these functions are general functions. An
example:

 - check_imbalance: Check for unbalanced reactions.
"""
import io
from pathlib import Path
from re import match
from typing import Any, Generator, Iterable, Iterator, Optional, TextIO

import cobra.core as cobra_core
from cobra import DictList, Reaction

import cobramod.error as cmod_error
from cobramod.debug import debug_log

ARROWS: dict[str, tuple[int, int]] = {
    "<=>": (-1000, 1000),
    "<->": (-1000, 1000),
    "==>": (0, 1000),
    "-->": (0, 1000),
    "<==": (-1000, 0),
    "<--": (-1000, 0),
}


def build_metabolite(
    identifier: str, formula: str, name: str, charge: float, compartment: str
) -> cobra_core.Metabolite:
    """
    Returns a basic :class:`cobra.Metabolite`. Creation of the metabolite
    writes in the log a with a DEBUG level.
    level.

    Args:
        identifier (str): Short name for Metabolite.
        formula (str): Chemical formula
        name (name): Long name for Metabolite
        charge (float): Charge of the metabolite
        compartment (str): Location

    Returns:
        Metabolite: new created metabolite
    """
    metabolite = cobra_core.Metabolite(
        id=identifier,
        formula=formula,
        name=name,
        charge=charge,
        compartment=compartment,
    )
    debug_log.debug(f'Curated metabolite "{identifier}" was created.')
    if not formula or formula.capitalize() == "X":
        debug_log.warning(
            f'Curated metabolite "{identifier} does not include a chemical '
            "formula"
        )
    return metabolite


def fix_name(name: str) -> str:
    """
    Returns new string, where hyphens are replaced to underscores
    """
    return name.replace("-", "_")


def check_imbalance(
    reaction: Reaction, stop_imbalance: bool, show_imbalance: bool
):
    """
    Verifies if given reaction is unbalanced in given model.

    Args:
        reaction (Reaction): Reaction object to examine.
        stop_imbalance (bool): If imbalance is found, stop process.
        show_imbalance (bool): If imbalance is found, show output.

    Raises:
        UnbalancedReaction: if given reaction is unbalanced.
    """
    dict_balance = reaction.check_mass_balance()

    # Will stop if True
    if dict_balance != {}:
        msg = (
            f'Reaction "{reaction.id}" unbalanced. Following atoms are '
            + f"affected. Please verify: {dict_balance}"
        )
        if stop_imbalance:
            raise cmod_error.UnbalancedReaction(
                identifier=reaction.id, dict_balance=str(dict_balance)
            )
        if show_imbalance:
            debug_log.warning(msg)


def get_key_dict(dictionary: dict, pattern: str) -> str:
    """
    From given pattern, return the first key of the dictionary that matches it
    or is a sub-string of it. It will raise a Warning if nothing found.
    """
    for key in dictionary.keys():
        if match(pattern=pattern, string=key):
            return str(key)
    raise cmod_error.PatternNotFound(
        "No pattern was found for given dictionary."
    )


def read_lines(f: TextIO) -> Iterator:
    """
    Reads Text I/O and returns iterator of line that are not comments nor
    blanks spaces
    """
    for line in f:
        line = line.strip()
        if line.startswith("#"):
            continue
        if not line:  # blank
            continue
        yield line


def create_replacement(filename: Path) -> dict:
    """
    Creates a dictionary build from given file. Key are the first word until
    a colon ':' is found. The rest represents the value
    """
    replace_dict = dict()
    with open(file=str(filename), mode="r") as f:
        for line in read_lines(f=f):
            key, value = line.rsplit(sep=":")
            replace_dict[key.strip().rstrip()] = value.strip().rstrip()
    return replace_dict


def compare_type(first: Any, second: Any):
    """
    Returns True is both objects are the same type, else raise TypeError.
    """
    if type(first) == type(second):
        return True
    else:
        raise TypeError("Given objects are not the same type.")


def compare_DictList(first: DictList, second: DictList) -> Generator:
    """
    Yields item identifiers from the first DictList that are not in the second
    DictList
    """
    compare_type(first=first[0], second=second[0])
    for item in first:
        if not second.has_id(id=item.id):
            yield item.id


def write_to_file(sequences: Iterable, filename: Path):
    """
    Writes to given file all items from the iterable in separate lines.
    """
    with open(file=str(filename), mode="w+") as f:
        f.writelines(line + "\n" for line in sequences)


def find_intersection(
    dictlist: cobra_core.DictList, query: dict[str, str], revert: bool
) -> Optional[str]:
    """
    Return the first item from the intersection of a DictList and the values of
    a dictionary. The identifiers from the DictList can be reverted to their
    original. Returns None if no intersection is found
    """
    # Format
    if revert:
        dict_set: set[str] = {
            item.id[:-2].replace("_", "-") for item in dictlist
        }
    # No format
    else:
        dict_set = {item.id for item in dictlist}
    # Error if nothing to pop, or empty Model
    try:
        common = dict_set.intersection(set(query.values()))
        return common.pop()
    except KeyError:
        return None


def is_compound(string: str) -> bool:
    try:
        float(string)

    except ValueError:
        if string[0] not in ["<", "-", "+"]:
            return True
    return False


def get_arrow_position(reaction_str: str) -> int:
    position = -1
    for item in ARROWS:
        position = reaction_str.find(item)

        if position != -1:
            break
    return position


def is_transport(reaction_str: str) -> bool:
    """
    Check if given reaction includes same participant in both sides,
    meaning that there is a tranport reaction. Prefix is not taken into
    consideration.
    """
    position = get_arrow_position(reaction_str)

    reactants = {
        item[2:].rstrip().strip()
        for item in reaction_str[:position].rstrip().strip().split(" ")
        if is_compound(item)
    }
    products = {
        item[2:].rstrip().strip()
        for item in reaction_str[position + 3 :].rstrip().strip().split(" ")
        if is_compound(item)
    }

    return len(reactants.intersection(products)) > 0


def convert_to_transport(reaction_str: str, compartment: str) -> str:
    """Modifies the reaction string such that the left side of the equation
    comes from the extracelullar compartment"""
    position = get_arrow_position(reaction_str)

    reactants = reaction_str[:position]
    products = reaction_str[position:]

    reactants = reactants.replace(f"_{compartment}", "_e")

    return reactants + products


def confirm_metabolite(
    model: cobra_core.Model, metabolites: list[cobra_core.Metabolite]
):
    """
    Checks if given metabolites are already in the model. If not, they will be
    added into the model.

    Args:
        model (Model): Model to extend.
        metabolites (List[Metabolites]): List with Metabolites.
    """
    # A Loop in necessary to log the skipped metabolites.
    for member in metabolites:
        found = model.metabolites.has_id(member.id)

        if not found:
            model.add_metabolites(metabolite_list=member)
            debug_log.info(f'Metabolite "{member.id}" was added to model.')
            continue

        msg = (
            f'Metabolite "{member.id}" is already present in the model. '
            "Skipping addition."
        )
        debug_log.warning(msg=msg)


def confirm_sink(model: cobra_core.Model, identifier: str):
    """
    Creates sink reactions for the metabolites that participate in given
    reaction. They can only be created if the corresponding metabolite only
    participates in one reaction. Otherwise, the metabolite can be somehow
    turnover
    """
    reaction = model.reactions.get_by_id(identifier)
    if not isinstance(reaction, cobra_core.Reaction):
        raise TypeError("Given object is not a COBRApy Reaction")

    metabolites: list[cobra_core.Metabolite] = (
        reaction.products + reaction.reactants
    )

    for metabolite in metabolites:
        amount = len(metabolite.reactions)

        if amount <= 1:
            # Warning is called in _inform_sink
            model.add_boundary(metabolite=metabolite, type="sink")


def add_reactions_to_model(
    model: cobra_core.Model, reactions: list[cobra_core.Reaction]
):
    """
    Check function that adds Reactions to given model if it does not
    contain the reaction. It logs the skipped reactions.
    """
    for member in reactions:
        if model.reactions.has_id(member.id):
            msg = (
                f'Reaction "{member.id}" is already present in the model. '
                "Skipping additions."
            )
            debug_log.warning(msg=msg)
            continue

        model.add_reactions([member])
        debug_log.info(f'Reaction "{member.id}" was added to model.')


def reaction_is_minimize(model: cobra_core.Model, identifier: str) -> bool:
    """
    Return whether given reaction should be minimized when optimizing
    """
    reaction = model.reactions.get_by_id(identifier)

    if not isinstance(reaction, cobra_core.Reaction):
        raise TypeError("Given object is not a COBRApy Reaction")

    lower = reaction.lower_bound
    upper = reaction.upper_bound

    if abs(lower) > upper:
        return True

    return False


def inform_new_sinks(model: cobra_core.Model, previous_sinks: set[str]):
    """
    Warns the user if sinks were created during the extension of the Model.
    i.e. Function compares a set with sink identifiers with the actual sinks
    in a model.
    """
    sinks: set[str] = {sink.id for sink in model.sinks if sink.id}.difference(
        previous_sinks
    )

    if sinks:
        for reaction in sinks:
            msg = (
                f'Auxiliary sink reaction for "{reaction}" created. Consider '
                "removing it and adding the synthesis reactions for the "
                "metabolite."
            )
            debug_log.warning(msg=msg)


def get_credentials(file: Path) -> tuple[str, str]:
    with open(file, "r") as f:
        user = f.readline().rstrip().strip()
        pwd = f.readline().rstrip().strip()

    return user, pwd


def kegg_info_to_version(info: str) -> str:
    with io.StringIO(info) as lines:
        for num, line in enumerate(lines, 1):
            if num == 2:
                return line[line.find("Release") + 8 :].rstrip()

    msg = (
        "Error determining the kegg version. "
        'Instead, "Undefined" is used as version.'
    )
    debug_log.warning(msg)
    return "Undefined"



#!/usr/bin/env python3
"""
Unit-test for data retrieval and Data version configuration. It checks that
the files are loaded and saved properly
"""
import logging
import shutil
import tempfile
import unittest
import warnings
from pathlib import Path
from unittest.mock import patch

import cobramod.retrieval as cmod_retrieval
import pandas as pd
from cobra import __version__ as cobra_version
from cobramod import __version__ as cmod_version
from cobramod.debug import change_to_debug
from cobramod.parsing.db_version import DataVersionConfigurator
from pandas._testing import assert_frame_equal, assert_series_equal

change_to_debug()
data_conf = DataVersionConfigurator()

dir_data = Path(__file__).resolve().parent.joinpath("data")

# If data is missing, then do not test. Data should always be the same
if not dir_data.exists():
    raise NotADirectoryError("Data for the test is missing")


class DataVersion(unittest.TestCase):
    directory: str

    @classmethod
    def setUp(cls):
        # removing metadata belonging to other tests that affect this one
        data_conf.database_version = None

        cls.directory = tempfile.mkdtemp()
        data_conf.ignore_db_versions = False
        cls.versions = pd.DataFrame.from_dict(
            {"orgid": ["bigg"], "version": ["1.0.0"]}
        )

        cls.versions.to_csv(
            path_or_buf=str(cls.directory + "/DatabaseVersions.csv"),
            index=False,
        )

    @classmethod
    def tearDown(cls):
        shutil.rmtree(cls.directory)
        # Removing metadata
        data_conf.database_version = None

    def test_get_database_version(self):
        database = data_conf.get_database_version(self.directory)
        assert_frame_equal(database, self.versions)

    def test_get_local_databases(self):
        databases = data_conf.get_local_databases(self.directory)

        expected_series = pd.Series(["bigg"])
        expected_series.name = "orgid"
        assert_series_equal(databases, expected_series)

    def test_set_database_version(self):
        data_conf.set_database_version(Path(self.directory), "pmn:META", "1.0")

        database = data_conf.get_database_version(self.directory)

        self.versions = pd.concat(
            [
                self.versions,
                pd.DataFrame({"orgid": ["pmn:META"], "version": ["1.0"]}),
            ],
            ignore_index=True,
        )

        assert_frame_equal(database, self.versions)

    def test_check_database_version(self):
        (Path(self.directory) / "DatabaseVersions.csv").unlink()

        # new database
        data_conf.check_database_version(Path(self.directory), "bigg", "1.0.0")

        database = data_conf.get_database_version(self.directory)
        assert_frame_equal(database, self.versions)

        # database with correct version

        with warnings.catch_warnings(record=True) as warning_list:
            data_conf.check_database_version(
                Path(self.directory), "bigg", "1.0.0"
            )

        self.assertEqual(len(warning_list), 0)

        database = data_conf.get_database_version(self.directory)
        assert_frame_equal(database, self.versions)

        # BUG: if running discovery, this fails otherwise
        data_conf.force_same_version = False

        # database with incorrect version (with user input)
        with patch("builtins.input", return_value="y") as mock:
            with self.assertLogs(level=logging.DEBUG) as cm:
                data_conf.check_database_version(
                    Path(self.directory), "bigg", "2.0.0"
                )
                self.assertIn(
                    "do not match. Remote has version 2.0.0", cm.output[0]
                )

            args, _ = mock.call_args
            assert args == ("Ignore version mismatch? (Y)es (N)o: ",)

        database = data_conf.get_database_version(self.directory)
        assert_frame_equal(database, self.versions)

        # database with incorrect version (without user input)
        data_conf.ignore_db_versions = True
        with self.assertLogs(level=logging.DEBUG) as cm:
            data_conf.check_database_version(
                Path(self.directory), "bigg", "2.0.0"
            )
            self.assertIn(
                "do not match. Remote has version 2.0.0", cm.output[0]
            )

        database = data_conf.get_database_version(self.directory)
        assert_frame_equal(database, self.versions)


class RetrievalTesting(unittest.TestCase):
    def test_get_data(self):
        dir_data.joinpath("META", "WATER.xml").unlink(True)
        self.assertFalse(dir_data.joinpath("META", "WATER.xml").exists())

        # Simple retrieval from Biocyc (metabolite)
        test_data = cmod_retrieval.get_data(
            directory=dir_data, database="META", identifier="WATER"
        )
        self.assertEqual(test_data.path, dir_data.joinpath("META", "WATER.xml"))
        self.assertTrue(test_data.path.exists())

        # simple retrieval from Biocyc (reaction)
        test_data = cmod_retrieval.get_data(
            directory=dir_data, database="ARA", identifier="GLYOXII-RXN"
        )
        self.assertEqual(
            test_data.path, dir_data.joinpath("ARA", "GLYOXII-RXN.xml")
        )
        self.assertTrue(test_data.path.exists())
        self.assertTrue(
            dir_data.joinpath("ARA", "GENES", "GLYOXII-RXN_genes.xml").exists(),
        )

        # Simple retrieval from KEGG (metabolite)
        test_data = cmod_retrieval.get_data(
            directory=dir_data, database="KEGG", identifier="C00001"
        )
        self.assertTrue(test_data.path.exists())
        self.assertTrue(dir_data.joinpath("KEGG", "C00001.txt").exists())

        # Simple retrieval from KEGG (Reaction)
        test_data = cmod_retrieval.get_data(
            directory=dir_data,
            database="KEGG",
            identifier="R02736",
            genome="eco",
        )
        self.assertEqual(
            test_data.path, dir_data.joinpath("KEGG", "R02736.txt")
        )
        self.assertTrue(test_data.path.exists())
        self.assertTrue(
            dir_data.joinpath("KEGG", "GENES", "R02736_genes.txt").exists(),
        )

        test_data = cmod_retrieval.get_data(
            directory=dir_data, identifier="R08618", database="KEGG"
        )
        self.assertEqual(
            test_data.path, dir_data.joinpath("KEGG", "R08618.txt")
        )
        self.assertTrue(test_data.path.exists())
        self.assertTrue(
            dir_data.joinpath("KEGG", "GENES", "R08618_genes.txt").exists(),
        )

        # Simple retrieval from KEGG (Pathway)
        test_data = cmod_retrieval.get_data(
            directory=dir_data, database="KEGG", identifier="M00001"
        )
        self.assertEqual(
            test_data.path, dir_data.joinpath("KEGG", "M00001.txt")
        )
        self.assertTrue(test_data.path.exists())

        # Simple retrieval from BIGG (metabolite)
        test_data = cmod_retrieval.get_data(
            directory=dir_data,
            database="BIGG",
            identifier="accoa_c",
            model_id="e_coli_core",
        )
        self.assertEqual(
            test_data.path,
            dir_data.joinpath("BIGG", "e_coli_core", "accoa_c.json"),
        )
        self.assertTrue(test_data.path.exists())

        # Simple retrieval from BIGG (reaction)
        test_data = cmod_retrieval.get_data(
            directory=dir_data,
            database="BIGG",
            identifier="CS",
            model_id="e_coli_core",
        )
        self.assertEqual(
            test_data.path, dir_data.joinpath("BIGG", "e_coli_core", "CS.json")
        )
        self.assertTrue(test_data.path.exists())

        # Retrieval from PMN
        test_data = cmod_retrieval.get_data(
            directory=dir_data, database="pmn:CORN", identifier="RXN-11501"
        )
        self.assertEqual(
            test_data.path, dir_data.joinpath("PMN", "CORN", "RXN-11501.xml")
        )
        self.assertTrue(test_data.path.exists())
        self.assertTrue(
            dir_data.joinpath(
                "PMN", "CORN", "GENES", "RXN-11501_genes.xml"
            ).exists(),
        )


if __name__ == "__main__":
    print(f"CobraMod version: {cmod_version}")
    print(f"COBRApy version: {cobra_version}")

    unittest.main(verbosity=2, failfast=True)



#!/usr/bin/env python3
"""Unittest for module extension

this module test the behavior of multiple functions, which are responsible to
add metabolites and reactions into a metabolic model. This module can be
divided in to parts:

- CreatingSequences: Functions, that create the corresponding reactions as
sequences and their corresponding flux test.
- AddingPathways: Functions, that manage the addition of Pathways into the
metabolic models.
"""
import logging
import unittest
from pathlib import Path

import cobra.core as cobra_core
import cobramod.error as cmod_error
import cobramod.retrieval as cmod_retrieval
import requests
from cobra import __version__ as cobra_version
from cobramod import __version__ as cmod_version
from cobramod.core import extension as ex
from cobramod.core.creation import add_reactions
from cobramod.core.pathway import Pathway
from cobramod.debug import change_to_debug
from cobramod.parsing.db_version import DataVersionConfigurator
from cobramod.test import textbook, textbook_biocyc, textbook_kegg

NAME = "test_model"

change_to_debug()
data_conf = DataVersionConfigurator()
data_conf.force_same_version = True

dir_data = Path(__file__).resolve().parent.joinpath("data")
dir_input = Path(__file__).resolve().parent.joinpath("input")

# If data is missing, then do not test. Data should always be the same
if not dir_data.exists():
    raise NotADirectoryError("Data for the test is missing")


class CreatingSequences(unittest.TestCase):
    def test_create_reactions(self):
        # CASE: Simple Case Biocyc
        generator = ex.yield_reaction_from_list(
            sequence=[
                "OXALODECARB-RXN",
                "AROMATIC-L-AMINO-ACID-DECARBOXYLASE-RXN",
            ],
            compartment="c",
            directory=dir_data,
            database="ARA",
            replacement={},
            show_imbalance=False,
            stop_imbalance=False,
            model=cobra_core.Model(NAME),
            model_id="",
            genome=None,
        )
        self.assertIsInstance(obj=next(generator), cls=cobra_core.Reaction)

        # CASE: Simple case Kegg
        generator = ex.yield_reaction_from_list(
            sequence=["R00200", "R00114"],
            compartment="p",
            directory=dir_data,
            database="KEGG",
            replacement={},
            show_imbalance=False,
            stop_imbalance=False,
            model=cobra_core.Model(NAME),
            model_id="",
            genome="eco",
        )
        self.assertIsInstance(obj=next(generator), cls=cobra_core.Reaction)
        # CASE: Showing when unbalanced
        generator = ex.yield_reaction_from_list(
            sequence=["RXN-11414", "RXN-11422"],
            compartment="c",
            directory=dir_data,
            database="ARA",
            replacement={},
            show_imbalance=True,
            stop_imbalance=False,
            model=cobra_core.Model(NAME),
            model_id="",
            genome=None,
        )
        # RXN-11414 is unbalanced
        with self.assertLogs(level=logging.DEBUG) as cm:
            next(generator)
            self.assertIn("unbalanced", cm.output[-1])

        # CASE: Stopping when unbalanced
        generator = ex.yield_reaction_from_list(
            sequence=["RXN-11414", "RXN-11422"],
            compartment="c",
            directory=dir_data,
            database="ARA",
            show_imbalance=False,
            stop_imbalance=True,
            replacement={},
            model=cobra_core.Model(NAME),
            model_id="",
            genome=None,
        )
        self.assertRaises(cmod_error.UnbalancedReaction, next, generator)

    def test_test_non_zero_flux(self):
        # CASE: Single Regular reaction
        test_model = cobra_core.Model(NAME)

        add_reactions(
            model=test_model,
            obj=["RXN-2206, c"],
            directory=dir_data,
            database="ARA",
            replacement={},
            show_imbalance=False,
        )
        ex.test_non_zero_flux(model=test_model, reaction="RXN_2206_c")
        self.assertEqual(first=7, second=len(test_model.sinks))

        # CASE: minimun range not reached
        # INFO: this extra steps are needed in order to make it fail
        test_model = textbook_biocyc.copy()
        add_reactions(
            model=test_model,
            obj=[
                "Redox_ADP_ATP_p, Redox_ADP_ATP_p | ADP_p <-> ATP_p",
                "TRANS_Pi_cp, Transport Phosphate_cp | Pi_c <-> Pi_p",
                "TRANS_GLUTATHIONE_cp, Transport GLUTATHIONE_cp |"
                + " GLUTATHIONE_c <-> GLUTATHIONE_p",
                "GLUTATHIONE-SYN-RXN, p",
            ],
            database="ARA",
            directory=dir_data,
            show_imbalance=False,
        )

        # CASE: direction right to left
        test_model = cobra_core.Model(NAME)
        add_reactions(
            model=test_model,
            obj=["1.8.4.9-RXN, c"],
            directory=dir_data,
            database="ARA",
        )
        ex.test_non_zero_flux(model=test_model, reaction="1.8.4.9_RXN_c")
        self.assertEqual(first=6, second=len(test_model.sinks))


class AddingPathways(unittest.TestCase):
    """
    Test for functions related to the addition of Pathways and their
    visualizations.
    """

    def test_add_reactions_to_Pathway(self):
        # CASE: Normal usage (3 reactions)
        test_model = textbook_biocyc.copy()
        test_list = list(
            ex.yield_reaction_from_list(
                sequence=["RXN-2206", "RXN-11414", "RXN-11422"],
                compartment="c",
                directory=dir_data,
                database="ARA",
                replacement={},
                show_imbalance=False,
                stop_imbalance=False,
                model=test_model,
                model_id="",
                genome=None,
            )
        )
        test_group = Pathway("test_group")

        ex.add_reactions_to_Pathway(
            model=test_model,
            pathway=test_group,
            sequence=test_list,
            ignore_list=["WATER_c", "OXYGEN_MOLECULE_c"],
        )
        self.assertGreater(
            abs(test_model.slim_optimize(error_value=0)),
            0,  # type: ignore
        )
        self.assertEqual(len(test_group.members), 3)

        # CASE: reactions already in model
        test_model = textbook.copy()
        reactions = [
            reaction
            for reaction in test_model.reactions
            if reaction.id in ("GAPD", "PGK", "PGM")
        ]
        test_group = Pathway("test_group")

        ex.add_reactions_to_Pathway(
            model=test_model,
            pathway=test_group,
            sequence=reactions,
            ignore_list=[],
        )
        self.assertEqual(len(test_group.members), 3)

        # CASE: KEGG
        test_model = textbook_kegg.copy()
        test_list = list(
            ex.yield_reaction_from_list(
                sequence=["R00894", "R00497"],
                compartment="c",
                directory=dir_data,
                database="KEGG",
                replacement={},
                show_imbalance=False,
                stop_imbalance=False,
                model=test_model,
                model_id="",
                genome="ath",
            )
        )
        test_group = Pathway("test_group")

        ex.add_reactions_to_Pathway(
            model=test_model,
            pathway=test_group,
            sequence=test_list,
            ignore_list=["WATER_c", "OXYGEN_MOLECULE_c"],
        )
        self.assertGreater(abs(test_model.slim_optimize(error_value=0)), 0)
        self.assertEqual(len(test_group.members), 2)

    def test_add_pathway_from_dict(self):
        data = cmod_retrieval.get_data(
            identifier="M00118",
            directory=dir_data,
            # debug_level=10,
            database="KEGG",
        )
        # CASE 1: regular test with KEGG
        test_model = textbook_kegg.copy()
        ex.add_pathway_from_data(
            model=test_model,
            group=None,
            data=data,
            directory=dir_data,
            database="KEGG",
            compartment="c",
            avoid_list=[],
            replacement={},
            ignore_list=[],
            show_imbalance=False,
            stop_imbalance=False,
            model_id="",
            genome="hsa",
        )
        self.assertIn(
            member="M00118",
            container=[group.id for group in test_model.groups],
        )
        for item in ["2729", "2937"]:
            self.assertIn(
                member=item, container=[gene.id for gene in test_model.genes]
            )
        # CASE 2: Using another label
        test_model = textbook_kegg.copy()
        ex.add_pathway_from_data(
            model=test_model,
            data=data,
            directory=dir_data,
            database="KEGG",
            group="ALTERNATIVE",
            compartment="c",
            avoid_list=[],
            replacement={},
            ignore_list=[],
            show_imbalance=False,
            stop_imbalance=False,
            model_id="",
            genome="hsa",
        )
        self.assertIn(
            member="ALTERNATIVE",
            container=[group.id for group in test_model.groups],
        )

    def test_add_pathway_from_list(self):
        # CASE: regular test
        test_model = textbook_biocyc.copy()
        ex.add_pathway_from_strings(
            model=test_model,
            identifier="test_group",
            compartment="c",
            sequence=["R01063", "R09084"],
            database="KEGG",
            directory=dir_data,
            avoid_list=[],
            replacement={},
            ignore_list=[],
            show_imbalance=False,
            stop_imbalance=False,
            model_id="",
            genome="mba",
        )
        self.assertTrue(test_model.groups.has_id("test_group"))

        for item in ["Mbar_A2189", "Mbar_A3564"]:
            self.assertTrue(test_model.genes.has_id(item))

    def test_add_pathway_from_file(self):
        test_model = cobra_core.Model(NAME)

        file = dir_input.joinpath("reactions_identifiers.txt")

        ex.add_pathway_from_file(
            model=test_model,
            file=file,
            replacement={"Not_GLC": "GLC"},
            identifier="test_group",
            show_imbalance=False,
            stop_imbalance=False,
            model_id="",
            database="META",
            directory=dir_data,
            ignore_list=[],
            genome=None,
        )
        for reaction in (
            "RXN_14462_c",
            "ACETALD_DEHYDROG_RXN_c",
            "AMONITRO_RXN_c",
        ):
            self.assertIn(
                member=reaction,
                container=[reaction.id for reaction in test_model.reactions],
            )

    def test_add_pathway(self):
        # CASE: Regular Biocyc
        test_model = textbook_biocyc.copy()

        ex.add_pathway(
            model=test_model,
            pathway="PWY-1187",
            compartment="c",
            directory=dir_data,
            database="ARA",
            ignore_list=[],
            show_imbalance=False,
        )
        self.assertIn(
            member="RXN_11438_c",
            container=[reaction.id for reaction in test_model.reactions],
        )
        for item in ["AT1G18590", "AT1G74090"]:
            self.assertIn(
                member=item, container=[gene.id for gene in test_model.genes]
            )
        # CASE: stacking another pathways (independent from each other)
        ex.add_pathway(
            model=test_model,
            pathway="AMMOXID-PWY",
            compartment="c",
            directory=dir_data,
            database="META",
            show_imbalance=False,
        )
        self.assertGreater(abs(test_model.slim_optimize(error_value=0)), 0)
        sol = test_model.optimize()

        # Test for demands, all should have a flux
        for demand in test_model.demands:
            value = sol.fluxes[demand.id]
            if not isinstance(value, float):
                raise TypeError("Value is not float")

            self.assertGreater(abs(value), 0)
        self.assertEqual(first=len(test_model.groups), second=2)

        # CASE: using a simple sequence
        test_model = textbook_biocyc.copy()
        test_sequence = ["RXN-2206", "RXN-11414", "RXN-11422", "RXN-11430"]
        ex.add_pathway(
            model=test_model,
            compartment="c",
            pathway=test_sequence,
            ignore_list=[],
            database="ARA",
            directory=dir_data,
            show_imbalance=False,
        )
        self.assertGreater(a=test_model.slim_optimize(), b=0)
        self.assertIn(
            member="RXN_11422_c",
            container=[reaction.id for reaction in test_model.reactions],
        )
        # CASE: Behavior if reaction was already in model
        test_model = textbook_biocyc.copy()

        # Adding reactions
        test_sequence = ["RXN-2206", "RXN-11414"]
        ex.add_pathway(
            model=test_model,
            group="old_reactions",
            compartment="c",
            pathway=test_sequence,
            ignore_list=[],
            database="META",
            directory=dir_data,
            show_imbalance=False,
        )
        test_sequence = ["RXN-2206", "RXN-11414", "RXN-11422", "RXN-11430"]
        ex.add_pathway(
            model=test_model,
            compartment="c",
            pathway=test_sequence,
            ignore_list=[],
            database="ARA",
            directory=dir_data,
            show_imbalance=False,
        )
        self.assertGreater(a=test_model.slim_optimize(), b=0)

        test_group = test_model.groups.get_by_id("custom_group")
        if not isinstance(test_group, Pathway):
            raise TypeError("Given object is not a valid Cobramod Pathway")

        for identifier in ["RXN_2206_c", "RXN_11414_c", "RXN_11422_c"]:
            self.assertIn(
                member=identifier,
                container=[reaction.id for reaction in test_group.members],
            )
        # CASE: KEGG simple pathway
        test_model = textbook_kegg.copy()
        ex.add_pathway(
            model=test_model,
            pathway="M00118",
            database="KEGG",
            directory=dir_data,
            compartment="c",
            show_imbalance=False,
            genome="hsa",
        )
        self.assertGreater(a=test_model.slim_optimize(), b=0)
        self.assertIn(
            member="R00894_c",
            container=[reaction.id for reaction in test_model.reactions],
        )
        # CASE: KEGG lineal coplex pathway
        test_model = textbook_kegg.copy()
        ex.add_pathway(
            model=test_model,
            pathway="M00001",
            database="KEGG",
            directory=dir_data,
            compartment="c",
            ignore_list=[],
            avoid_list=[],
            show_imbalance=False,
            genome="hsa",
        )
        self.assertGreater(a=test_model.slim_optimize(), b=0)
        self.assertIn(
            member="R01063_c",
            container=[reaction.id for reaction in test_model.reactions],
        )
        self.assertIn(
            member="R01518_c",
            container=[reaction.id for reaction in test_model.reactions],
        )
        # CASE: Check for translations in pathways
        test_model = textbook_kegg.copy()
        ex.add_pathway(
            model=test_model,
            pathway="SALVADEHYPOX-PWY",
            database="ECO",
            directory=dir_data,
            compartment="c",
            show_imbalance=False,
        )
        self.assertGreater(a=test_model.slim_optimize(), b=0)
        test_reaction = test_model.reactions.get_by_id("ADENODEAMIN_RXN_c")

        if not isinstance(test_reaction, cobra_core.Reaction):
            raise TypeError("Given object is not a valid Cobramod Pathway")

        self.assertIn(
            member="C00080_c",
            container=[
                metabolite.id for metabolite in test_reaction.metabolites
            ],
        )
        # CASE: From sequence. Skipping already in Model. Stacking in one.
        test_model = textbook_kegg.copy()
        ex.add_pathway(
            model=test_model,
            pathway=["ACALD", "MALS"],
            database="BIGG",
            model_id="e_coli_core",
            directory=dir_data,
            compartment="c",
            show_imbalance=False,
            ignore_list=["R00228_c", "R00472_c"],
        )

        ex.add_pathway(
            model=test_model,
            pathway=["ADENODEAMIN-RXN"],
            database="ECO",
            directory=dir_data,
            compartment="c",
            show_imbalance=False,
        )
        test_reaction = test_model.reactions.get_by_id("ADENODEAMIN_RXN_c")
        if not isinstance(test_reaction, cobra_core.Reaction):
            raise TypeError("Given object is not a valid COBRApy Reaction")

        test_group = test_model.groups.get_by_id("custom_group")
        if not isinstance(test_group, Pathway):
            raise TypeError("Given object is not a valid Cobramod Pathway")

        self.assertEqual(first=len(test_group.members), second=3)
        self.assertGreater(a=test_model.slim_optimize(), b=0)
        self.assertNotIn(
            member="ACALD_c",
            container=[reaction.id for reaction in test_model.reactions],
        )
        self.assertNotIn(
            member="MALS_c",
            container=[reaction.id for reaction in test_model.reactions],
        )
        # WATER
        self.assertIn(
            member="C00001_c",
            container=[
                metabolite.id for metabolite in test_reaction.metabolites
            ],
        )
        # CASE: Testing avoid list
        test_model = textbook_kegg.copy()
        test_sequence = ["RXN-2206", "RXN-11414", "RXN-11422", "RXN-11430"]
        ex.add_pathway(
            model=test_model,
            pathway=test_sequence,
            database="ARA",
            directory=dir_data,
            compartment="c",
            show_imbalance=False,
            avoid_list=["RXN-2206"],
        )
        test_group = test_model.groups.get_by_id("custom_group")
        if not isinstance(test_group, Pathway):
            raise TypeError("Given object is not a valid Cobramod Pathway")

        self.assertEqual(first=len(test_group.members), second=3)
        self.assertEqual(first=len(test_group.graph), second=3)

        # CASE: Stacking pathways in one pathways
        test_sequence = ["RXN-11438", "RXN-2208", "RXN-2209", "RXN-2221"]
        ex.add_pathway(
            model=test_model,
            pathway=test_sequence,
            database="ARA",
            directory=dir_data,
            compartment="c",
            show_imbalance=False,
        )
        test_group = test_model.groups.get_by_id("custom_group")
        if not isinstance(test_group, Pathway):
            raise TypeError("Given object is not a valid Cobramod Pathway")

        self.assertEqual(first=len(test_group.members), second=7)
        self.assertEqual(first=len(test_group.graph), second=7)

        # CASE: Using replacement arguments to rename
        test_model = textbook_biocyc.copy()
        test_replacement = {
            "RXN-8052": "Reaction_8052",
            "RXN-2208": "Reaction_2208",
            "RXN-2223": "Reaction_2223",
        }
        ex.add_pathway(
            model=test_model,
            pathway="PWY-1187",
            compartment="c",
            directory=dir_data,
            database="ARA",
            replacement=test_replacement,
            show_imbalance=False,
        )
        test_group = test_model.groups.get_by_id("PWY-1187")
        if not isinstance(test_group, Pathway):
            raise TypeError("Given object is not a valid Cobramod Pathway")

        for meta in test_replacement.values():
            identifier = f"{meta}_c"
            self.assertIn(
                member=identifier,
                container=[reaction.id for reaction in test_model.reactions],
            )
            self.assertIn(
                member=identifier,
                container=[reaction.id for reaction in test_group.members],
            )

        # CASE: Using replacement arguments to replace
        test_model = textbook_biocyc.copy()
        test_sequence = ["RXN-2206", "RXN-11414", "RXN-11422", "RXN-11430"]

        ex.add_pathway(
            model=test_model,
            compartment="c",
            pathway=test_sequence,
            ignore_list=[],
            database="ARA",
            directory=dir_data,
            replacement={"RXN-2206": "ACALDt"},
            show_imbalance=False,
        )

        test_group = test_model.groups.get_by_id("custom_group")
        if not isinstance(test_group, Pathway):
            raise TypeError("Given object is not a valid Cobramod Pathway")

        self.assertIn(
            member="ACALDt_c",
            container=[reaction.id for reaction in test_group.members],
        )
        # CASE: Raising AttributeError
        test_model = textbook_biocyc.copy()

        self.assertRaises(
            requests.HTTPError,
            ex.add_pathway,
            model=test_model,
            compartment="c",
            pathway=dir_input.joinpath("reactions_identifiers.txt"),
            database="ARA",
            ignore_list=[],
            directory=dir_data,
            show_imbalance=False,
        )

        # CASE: Using Path
        test_model = textbook_biocyc.copy()

        ex.add_pathway(
            model=test_model,
            compartment="c",
            pathway=dir_input.joinpath("reactions_identifiers.txt"),
            database="META",
            ignore_list=[],
            directory=dir_data,
            show_imbalance=False,
        )

        test_group = test_model.groups.get_by_id("custom_group")
        if not isinstance(test_group, Pathway):
            raise TypeError("Given object is not a valid Cobramod Pathway")

        for reaction in (
            "RXN_14462_c",
            "ACETALD_DEHYDROG_RXN_c",
            "AMONITRO_RXN_c",
        ):
            self.assertIn(
                member=reaction,
                container=[reaction.id for reaction in test_group.members],
            )


if __name__ == "__main__":
    print(f"CobraMod version: {cmod_version}")
    print(f"COBRApy version: {cobra_version}")

    unittest.main(verbosity=2, failfast=True)



#!/usr/bin/env python3
"""Unittest for genes verification

In this module, the creation of genes for multiple function and their behavior
are checked
"""
import logging
import unittest
from pathlib import Path

import cobra.core as cobra_core
from cobra import __version__ as cobra_version
from cobramod import __version__ as cmod_version
from cobramod.core.creation import create_object
from cobramod.debug import change_to_debug
from cobramod.parsing.db_version import DataVersionConfigurator

change_to_debug()
data_conf = DataVersionConfigurator()
data_conf.force_same_version = True

dir_data = Path(__file__).resolve().parent.joinpath("data")

# If data is missing, then do not test. Data should always be the same
if not dir_data.exists():
    raise NotADirectoryError("Data for the test is missing")


class TestComplexGenes(unittest.TestCase):
    def test_KEGG_genome(self):
        # CASE: catch warning logs
        with self.assertLogs(level=logging.DEBUG) as cm:
            create_object(
                identifier="R02736",
                directory=dir_data,
                compartment="c",
                database="KEGG",
                genome="fake",
            )
            self.assertIn(
                "abbreviation as a specie. No genes will be added.",
                cm.output[0],
            )
        test_reaction = create_object(
            identifier="R02736",
            directory=dir_data,
            compartment="c",
            database="KEGG",
            genome="fake",
        )
        if not isinstance(test_reaction, cobra_core.Reaction):
            raise TypeError("Given object is not a valid COBRApy Reaction")

        self.assertEqual(first=len(test_reaction.genes), second=0)

        # CASE: catch warning. No genome and thus, no genes
        with self.assertLogs(level=logging.DEBUG) as cm:
            create_object(
                identifier="R02736",
                directory=dir_data,
                compartment="c",
                database="KEGG",
            )
            self.assertIn(
                "Nothing was specified in argument ",
                cm.output[-1],
            )
        test_reaction = create_object(
            identifier="R02736",
            directory=dir_data,
            compartment="c",
            database="KEGG",
        )
        if not isinstance(test_reaction, cobra_core.Reaction):
            raise TypeError("Given object is not a valid COBRApy Reaction")

        self.assertEqual(first=len(test_reaction.genes), second=0)

        # CASE: regular case
        test_reaction = create_object(
            identifier="R02736",
            directory=dir_data,
            compartment="c",
            database="KEGG",
            genome="hsa",
        )
        if not isinstance(test_reaction, cobra_core.Reaction):
            raise TypeError("Given object is not a valid COBRApy Reaction")

        self.assertEqual(first=len(test_reaction.genes), second=2)


if __name__ == "__main__":
    print(f"CobraMod version: {cmod_version}")
    print(f"COBRApy version: {cobra_version}")

    unittest.main(verbosity=2, failfast=True)



"""Sub-package visualization

This package handles the creation and parsing of a JSON that can be used for
the package Escher. The main module converter.py has the class JsonDictionary,
which is the base for the Escher-visualization. The module items.py, is a
pythonic-representation of the JSON-objects from the Escher-schema, which
includes:

- Reactions
- Segments
- Nodes

Moreover, there is a debug module, which helps debugging the objects and the
initialization of the module. Please refer to each individual module for more
in-depth information.
"""



#!/usr/bin/env python3
"""Unittest for module creation

In this test, multiple single functions for the creation of reactions and
metabolites are tested. Important is the independency of the functions, unless
they are part of the API such as add_metabolites. Therefore the modules is
separated in:

- SimpleFunctions: Creation of objects
- ComplexFunctions: Functions, that uses multiple simple functions.
"""
import unittest
from logging import DEBUG
from pathlib import Path

import cobra.core as cobra_core
import cobramod.retrieval as cmod_retrieval
import requests
from cobra import __version__ as cobra_version
from cobramod import __version__ as cmod_version
from cobramod.core import creation as cr
from cobramod.debug import debug_log
from cobramod.error import NoDelimiter, WrongSyntax
from cobramod.parsing.db_version import DataVersionConfigurator
from cobramod.test import textbook_kegg

NAME = "test_model"

debug_log.setLevel(DEBUG)
data_conf = DataVersionConfigurator()
data_conf.force_same_version = True

dir_data = Path(__file__).resolve().parent.joinpath("data")
dir_input = Path(__file__).resolve().parent.joinpath("input")

# If data is missing, then do not test. Data should always be the same
if not dir_data.exists():
    raise NotADirectoryError("Data for the test is missing")


class SimpleFunctions(unittest.TestCase):
    """
    Test for simple test such as creating metabolites from string or from
    files.
    """

    def test_find_replacements(self):
        replace_dict = {
            "ACETALD-DEHYDROG-RXN": "R00228_c",
            "WATER_c": "C00001_c",
            "WATER_d": "C00001_d",
        }
        # CASE 1: Reaction
        test_replacement = cr.find_replacements(
            identifier="WATER_c",
            obj_type="metabolites",
            model=textbook_kegg,
            replace_dict=replace_dict,
        )
        self.assertIsInstance(obj=test_replacement, cls=cobra_core.Metabolite)
        # CASE: String
        test_replacement = cr.find_replacements(
            identifier="WATER_d",
            obj_type="metabolites",
            model=textbook_kegg,
            replace_dict=replace_dict,
        )
        self.assertIsNone(test_replacement)
        # CASE : String
        test_replacement = cr.find_replacements(
            identifier="ACETALD-DEHYDROG-RXN",
            obj_type="reactions",
            model=textbook_kegg,
            replace_dict=replace_dict,
        )
        self.assertIsInstance(obj=test_replacement, cls=cobra_core.Reaction)

    def test_metabolite_from_string(self):
        # CASE 1: Correct input
        test_string = "MALTOSE_b, MALTOSE[b], b, C12H22O11, 0"
        test_metabolite = cr.metabolite_from_string(
            line_string=test_string,
            model=cobra_core.Model(NAME),
            replacement={},
        )
        # Checking that new meta is in Model
        self.assertIsInstance(obj=test_metabolite, cls=cobra_core.Metabolite)
        self.assertEqual(first=test_metabolite.id, second="MALTOSE_b")
        self.assertEqual(first=test_metabolite.name, second="MALTOSE[b]")
        self.assertEqual(first=test_metabolite.charge, second=0)
        self.assertEqual(first=test_metabolite.formula, second="C12H22O11")

    def test_metabolite_from_data(self):
        # CASE: regular META
        data = cmod_retrieval.get_data(
            directory=dir_data,
            identifier="HOMOMETHIONINE",
            database="META",
            # debug_level=10,
        )
        test_metabolite = cr.metabolite_from_data(data, compartment="c")

        self.assertIsInstance(obj=test_metabolite, cls=cobra_core.Metabolite)
        self.assertEqual(first=test_metabolite.id, second="HOMOMETHIONINE_c")
        self.assertEqual(first=test_metabolite.name, second="L-homomethionine")
        self.assertEqual(first=test_metabolite.formula, second="C6H13N1O2S1")
        # CASE: Find translation
        data = cmod_retrieval.get_data(
            directory=dir_data,
            identifier="WATER",
            database="META",
            # debug_level=10,
        )
        test_metabolite = cr.metabolite_from_data(
            data, compartment="c", model=textbook_kegg
        )

        self.assertEqual(first=test_metabolite.id, second="C00001_c")
        # CASE: Find translation (BIGG)
        data = cmod_retrieval.get_data(
            directory=dir_data,
            identifier="h2o",
            database="BIGG",
            # debug_level=10,
            model_id="universal",
        )
        test_metabolite = cr.metabolite_from_data(
            data, compartment="c", model=textbook_kegg
        )
        self.assertEqual(first=test_metabolite.id, second="C00001_c")

    def test_convert_string_metabolite(self):
        # CASE 1: retrieval from META
        test_metabolite = cr.convert_string_metabolite(
            line="HOMOMETHIONINE, c",
            model=cobra_core.Model(NAME),
            directory=dir_data,
            database="META",
            replacement={},
            model_id=None,
        )
        self.assertIsInstance(obj=test_metabolite, cls=cobra_core.Metabolite)
        self.assertEqual(first=test_metabolite.id, second="HOMOMETHIONINE_c")
        self.assertEqual(first=test_metabolite.name, second="L-homomethionine")
        self.assertEqual(first=test_metabolite.formula, second="C6H13N1O2S1")
        self.assertEqual(first=test_metabolite.compartment, second="c")
        # CASE: custom metabolite
        test_metabolite = cr.convert_string_metabolite(
            line="MALTOSE_b, MALTOSE[b], b, C12H22O11, 0",
            model=cobra_core.Model(NAME),
            replacement={},
            directory=dir_data,
            database=None,
            model_id=None,
        )
        self.assertEqual(first=test_metabolite.id, second="MALTOSE_b")
        self.assertEqual(first=test_metabolite.name, second="MALTOSE[b]")
        self.assertEqual(first=test_metabolite.charge, second=0)
        self.assertEqual(first=test_metabolite.formula, second="C12H22O11")
        # CASE: custom metabolite with replacement
        test_metabolite = cr.convert_string_metabolite(
            line="MALTOSE_b, MALTOSE[c], b, C12H22O11, 0",
            model=cobra_core.Model(NAME),
            replacement={"MALTOSE_b": "MALTOSE_c"},
            directory=dir_data,
            database=None,
            model_id=None,
        )
        self.assertEqual(first=test_metabolite.id, second="MALTOSE_c")
        self.assertEqual(first=test_metabolite.name, second="MALTOSE[c]")
        self.assertEqual(first=test_metabolite.charge, second=0)
        self.assertEqual(first=test_metabolite.formula, second="C12H22O11")

    def test_get_file_metabolites(self):
        test_model = cobra_core.Model(NAME)
        # CASE: Testing if file is not found
        self.assertRaises(
            FileNotFoundError,
            cr.get_file_metabolites,
            test_model,
            Path.cwd().joinpath("no_file"),
            replacement={},
            directory=dir_data,
            database=None,
            model_id=None,
        )
        # CASE: Metabolite is not found (or misspelled)
        self.assertRaises(
            requests.HTTPError,
            cr.get_file_metabolites,
            test_model,
            dir_input.joinpath("metabolites_02_misspelled.txt"),
            replacement={},
            directory=dir_data,
            database="META",
            model_id=None,
        )
        # CASE: Bad format (e. g. charge is missing).
        self.assertRaises(
            WrongSyntax,
            cr.get_file_metabolites,
            model=test_model,
            filename=dir_input.joinpath("metabolites_03_badFormat.txt"),
            directory=dir_data,
            database="META",
            replacement={},
            model_id=None,
        )
        # CASE: Normal input
        test_list = cr.get_file_metabolites(
            model=test_model,
            filename=dir_input.joinpath("metabolites_01_normal.txt"),
            directory=dir_data,
            database="META",
            replacement={},
            model_id=None,
        )
        self.assertEqual(len(test_list), 2)

        # Both Names in metabolites
        test_names = ["HOMOMETHIONINE_c", "MALTOSE_b"]
        self.assertListEqual(
            list1=[member.id for member in test_list], list2=test_names
        )

    def test_get_reaction(self):
        # CASE: Regular Biocyc reaction
        data = cmod_retrieval.get_data(
            directory=dir_data,
            identifier="OXALODECARB-RXN",
            database="ARA",
            # debug_level=10,
        )
        test_reaction = cr.get_reaction(
            data=data,
            compartment="c",
            replacement={},
            model=cobra_core.Model(NAME),
            show_imbalance=True,
            stop_imbalance=False,
        )
        self.assertIsInstance(obj=test_reaction, cls=cobra_core.Reaction)
        self.assertTupleEqual(tuple1=test_reaction.bounds, tuple2=(0, 1000))
        test_list = [
            "PYRUVATE_c",
            "CARBON_DIOXIDE_c",
            "OXALACETIC_ACID_c",
            "PROTON_c",
        ]
        self.assertCountEqual(
            [meta.id for meta in test_reaction.metabolites],
            test_list,
        )
        # CASE: Regular KEGG reaction in compartment "p". Irreversible
        data = cmod_retrieval.get_data(
            directory=dir_data,
            identifier="R02736",
            database="KEGG",
            # debug_level=10,
            # genome="hsa",
        )
        test_reaction = cr.get_reaction(
            data=data,
            compartment="p",
            replacement={},
            model=cobra_core.Model(NAME),
            show_imbalance=True,
            stop_imbalance=False,
        )
        self.assertIsInstance(obj=test_reaction, cls=cobra_core.Reaction)
        # NOTE: dropping support from comments
        self.assertTupleEqual(tuple1=test_reaction.bounds, tuple2=(-1000, 1000))
        test_list = [
            "C01172_p",
            "C00006_p",
            "C01236_p",
            "C00005_p",
            "C00080_p",
        ]
        self.assertCountEqual(
            [meta.id for meta in test_reaction.metabolites],
            test_list,
        )
        # CASE: Reactions with different coefficients
        data = cmod_retrieval.get_data(
            directory=dir_data,
            identifier="R00114",
            database="KEGG",
            # debug_level=10,
            # genome="eco",
        )
        test_reaction = cr.get_reaction(
            data=data,
            compartment="p",
            replacement={},
            model=cobra_core.Model(NAME),
            show_imbalance=True,
            stop_imbalance=False,
        )
        self.assertIsInstance(obj=test_reaction, cls=cobra_core.Reaction)
        self.assertTupleEqual(tuple1=test_reaction.bounds, tuple2=(-1000, 1000))
        test_list = [
            "C00025_p",
            "C00006_p",
            "C00064_p",
            "C00026_p",
            "C00080_p",
            "C00005_p",
        ]
        self.assertListEqual(
            list1=sorted([meta.id for meta in test_reaction.metabolites]),
            list2=sorted(test_list),
        )
        self.assertEqual(
            first=test_reaction.get_coefficient("C00025_p"), second=-2
        )
        # CASE: transport reaction, simple
        data = cmod_retrieval.get_data(
            directory=dir_data,
            identifier="TRANS-RXN0-574",
            database="META",
            # debug_level=10,
        )
        test_reaction = cr.get_reaction(
            data=data,
            compartment="p",
            replacement={},
            model=cobra_core.Model(NAME),
            show_imbalance=True,
            stop_imbalance=False,
        )
        self.assertIn(
            member="Glucopyranose_p",
            container=[item.id for item in test_reaction.metabolites],
        )
        self.assertIn(
            member="Glucopyranose_e",
            container=[item.id for item in test_reaction.metabolites],
        )
        # CASE: transport reaction, complex
        data = cmod_retrieval.get_data(
            directory=dir_data,
            identifier="ABC-56-RXN",
            database="META",
            # debug_level=10,
        )
        test_reaction = cr.get_reaction(
            data=data,
            compartment="p",
            replacement={},
            model=cobra_core.Model(NAME),
            show_imbalance=True,
            stop_imbalance=False,
        )
        self.assertIn(
            member="Aliphatic_Sulfonates_e",
            container=[item.id for item in test_reaction.metabolites],
        )
        self.assertIn(
            member="Aliphatic_Sulfonates_p",
            container=[item.id for item in test_reaction.metabolites],
        )
        self.assertEqual(
            first=test_reaction.get_coefficient("Aliphatic_Sulfonates_e"),
            second=-1,
        )
        # CASE: Retrieve reaction with translated equivalents
        data = cmod_retrieval.get_data(
            directory=dir_data,
            identifier="ACALD",
            database="BIGG",
            # debug_level=10,
            model_id="e_coli_core",
        )
        test_reaction = cr.get_reaction(
            data=data,
            compartment="c",
            replacement={},
            model=textbook_kegg,
            show_imbalance=True,
            stop_imbalance=False,
        )
        self.assertEqual(first=test_reaction.id, second="R00228_c")
        # CASE: Retrieve reaction with translated equivalents. Check for
        # metabolites.
        data = cmod_retrieval.get_data(
            directory=dir_data,
            identifier="ADENODEAMIN-RXN",
            database="META",
            # debug_level=10,
        )
        test_reaction = cr.get_reaction(
            data=data,
            compartment="c",
            replacement={},
            model=textbook_kegg,
            show_imbalance=True,
            stop_imbalance=False,
        )
        self.assertEqual(first=test_reaction.id, second="ADENODEAMIN_RXN_c")
        # WATER
        self.assertIn(
            member="C00001_c",
            container=[
                metabolite.id for metabolite in test_reaction.metabolites
            ],
        )
        # PROTON
        self.assertIn(
            member="C00080_c",
            container=[
                metabolite.id for metabolite in test_reaction.metabolites
            ],
        )

    def test_convert_string_reaction(self):
        # CASE: using delimiter, compartment is cytosol
        test_model = cobra_core.Model(NAME)
        test_line = (
            "RXN_17742_c, RXN_17742_c | "
            "1 Oxidized-ferredoxins_c <-> 1 Reduced-ferredoxins_c "
        )
        test_reaction = cr.string_to_reaction(
            line=test_line,
            model=test_model,
            directory=dir_data,
            database="META",
            stop_imbalance=False,
            show_imbalance=True,
            replacement={},
            model_id=None,
            genome=None,
        )
        self.assertEqual(first="RXN_17742_c", second=test_reaction.id)
        for metabolite in ["Oxidized_ferredoxins_c", "Reduced_ferredoxins_c"]:
            self.assertIn(
                member=metabolite,
                container=[meta.id for meta in test_reaction.metabolites],
            )
        # CASE: No delimiter
        test_model = cobra_core.Model(NAME)
        test_line = "RXN-14462, c"
        test_reaction = cr.string_to_reaction(
            line=test_line,
            model=test_model,
            directory=dir_data,
            database="META",
            stop_imbalance=False,
            show_imbalance=True,
            replacement={},
            model_id=None,
            genome=None,
        )
        self.assertEqual(first="RXN_14462_c", second=test_reaction.id)

        # CASE: No delimiter, compartment p
        test_model = cobra_core.Model(NAME)
        test_line = "RXN-14462, p"
        test_reaction = cr.string_to_reaction(
            line=test_line,
            model=test_model,
            directory=dir_data,
            database="GCF_000020025",
            stop_imbalance=False,
            show_imbalance=True,
            replacement={},
            model_id=None,
            genome=None,
        )
        self.assertEqual(first="RXN_14462_p", second=test_reaction.id)

        # CASE: 100% Custom metabolite
        test_model = cobra_core.Model(NAME)
        test_line = "CUSTOM_rxn1_p, Custom_reaction | Meta_A_p --> Meta_B_p"
        test_reaction = cr.string_to_reaction(
            line=test_line,
            model=test_model,
            directory=dir_data,
            database="META",
            stop_imbalance=False,
            show_imbalance=True,
            replacement={},
            model_id=None,
            genome=None,
        )
        self.assertEqual(first="CUSTOM_rxn1_p", second=test_reaction.id)
        for metabolite in ["Meta_A_p", "Meta_B_p"]:
            self.assertIn(
                member=metabolite,
                container=[meta.id for meta in test_reaction.metabolites],
            )

    def test_reaction_from_string(self):
        # CASE: wrong format, no delimiter
        self.assertRaises(
            NoDelimiter,
            cr.reaction_from_string,
            line_string="GLC_cb, GLC_cb GLC_c <-> GLC_b",
            directory=dir_data,
            database="META",
            stop_imbalance=False,
            show_imbalance=True,
            model_id=None,
            replacement={},
        )

        # CASE: Normal, ID and name differ
        test_reaction = cr.reaction_from_string(
            line_string="GLC_cb, Glucose Transport| 2 GLC_c <=> GLC_b",
            directory=dir_data,
            database="META",
            stop_imbalance=False,
            show_imbalance=False,
            replacement={},
            model_id=None,
        )
        self.assertEqual(first=test_reaction.id, second="GLC_cb")
        self.assertEqual(first=test_reaction.name, second="Glucose Transport")
        self.assertIsInstance(obj=test_reaction, cls=cobra_core.Reaction)

        test_list = ["GLC_c", "GLC_b"]
        for metabolite in test_list:
            self.assertIn(
                member=metabolite,
                container=[meta.id for meta in test_reaction.metabolites],
            )
        self.assertEqual(
            first=-2, second=test_reaction.get_coefficient("GLC_c")
        )
        self.assertEqual(first=1, second=test_reaction.get_coefficient("GLC_b"))

        # CASE: Larger reaction
        test_reaction = cr.reaction_from_string(
            line_string="GLC_cb, Glucose Transport|"
            + "4 GLT_c + 2 GLC_c <=> GLC_b + 2 GLT_b",
            directory=dir_data,
            database="META",
            stop_imbalance=False,
            show_imbalance=False,
            replacement={},
            model_id=None,
        )
        self.assertEqual(
            first=-4, second=test_reaction.get_coefficient("GLT_c")
        )
        self.assertEqual(first=2, second=test_reaction.get_coefficient("GLT_b"))

    def test_get_file_reactions(self):
        test_model = cobra_core.Model(NAME)
        test_list = cr.get_file_reactions(
            model=test_model,
            filename=dir_input.joinpath("reactions_normal.txt"),
            directory=dir_data,
            database="META",
            replacement={},
            model_id=None,
            genome=None,
            stop_imbalance=False,
            show_imbalance=True,
        )
        test_names = ["GLC_cb", "RXN_14462_p"]
        self.assertListEqual(
            list1=test_names, list2=[reaction.id for reaction in test_list]
        )


class ComplexFunctions(unittest.TestCase):
    # @classmethod
    # def setUpClass(cls):
    #     BaseParser.ignore_db_versions = True

    def test_create_object(self):
        # CASE: metabolite from MetaCyc
        test_object = cr.create_object(
            identifier="GLC",
            directory=dir_data,
            compartment="c",
            database="META",
        )
        self.assertIsInstance(obj=test_object, cls=cobra_core.Metabolite)
        self.assertEqual(first=getattr(test_object, "compartment"), second="c")
        # CASE: metabolite from KEGG
        test_object = cr.create_object(
            identifier="C00026",
            directory=dir_data,
            compartment="p",
            database="KEGG",
        )
        self.assertIsInstance(obj=test_object, cls=cobra_core.Metabolite)
        self.assertEqual(first=getattr(test_object, "compartment"), second="p")
        # CASE: reaction from MetaCyc
        test_object = cr.create_object(
            identifier="2.6.1.58-RXN",
            directory=dir_data,
            compartment="p",
            database="META",
        )
        self.assertIsInstance(obj=test_object, cls=cobra_core.Reaction)
        self.assertIn(
            member="p", container=getattr(test_object, "compartments")
        )
        # CASE: reaction from KEGG
        test_object = cr.create_object(
            identifier="R02736",
            directory=dir_data,
            compartment="c",
            database="KEGG",
            genome="hsa",
        )
        self.assertIsInstance(obj=test_object, cls=cobra_core.Reaction)
        self.assertIn(
            member="c", container=getattr(test_object, "compartments")
        )
        self.assertCountEqual(
            first=[gene.id for gene in getattr(test_object, "genes")],
            second=["9563", "2539"],
        )

        # CASE: Reaction of Biocyc which could be a pathway as well
        test_object = cr.create_object(
            identifier="AMONITRO-RXN",
            directory=dir_data,
            compartment="c",
            database="META",
            show_imbalance=False,
        )
        self.assertIsInstance(obj=test_object, cls=cobra_core.Reaction)
        # CASE: Metabolites has correct formula
        test_object = cr.create_object(
            identifier="C00404",
            directory=dir_data,
            compartment="c",
            database="KEGG",
            show_imbalance=False,
        )
        if not isinstance(test_object, cobra_core.Metabolite):
            raise TypeError("Not a valid COBRApy object")
        if test_object.formula is None:
            raise TypeError("Formula is empty!")

        self.assertEqual(test_object.formula.find("("), -1)

        # CASE: Replacing names (Metabolite)
        test_replacement = {
            "GLC": "GLUCOSE",
            "AMONITRO-RXN": "monooxygenase",
            "Donor-H2": "Donor_new_name",
        }

        test_object = cr.create_object(
            identifier="GLC",
            directory=dir_data,
            compartment="c",
            database="META",
            replacement=test_replacement,
        )
        self.assertIsInstance(obj=test_object, cls=cobra_core.Metabolite)
        self.assertEqual(first=getattr(test_object, "compartment"), second="c")
        self.assertEqual(first=getattr(test_object, "id"), second="GLUCOSE_c")

        # CASE: Replacing names (Reaction)
        test_object = cr.create_object(
            identifier="AMONITRO-RXN",
            directory=dir_data,
            compartment="c",
            database="META",
            replacement=test_replacement,
        )
        self.assertIsInstance(obj=test_object, cls=cobra_core.Reaction)
        self.assertCountEqual(
            first=getattr(test_object, "compartments"), second={"c"}
        )
        self.assertEqual(
            first=getattr(test_object, "id"), second="monooxygenase_c"
        )
        self.assertIn(
            member="Donor_new_name_c",
            container=[meta.id for meta in getattr(test_object, "metabolites")],
        )

    def test_add_metabolites(self):
        # CASE: From path
        test_model = cobra_core.Model(NAME)
        cr.add_metabolites(
            model=test_model,
            obj=dir_input.joinpath("metabolites_01_normal.txt"),
            directory=dir_data,
            database="META",
        )
        # CASE: path as str
        cr.add_metabolites(
            model=test_model,
            obj=str(dir_input.joinpath("metabolites_01_normal.txt")),
            directory=dir_data,
            database="META",
        )
        test_names = ["HOMOMETHIONINE_c", "MALTOSE_b"]
        self.assertListEqual(
            list1=[member.id for member in test_model.metabolites],
            list2=test_names,
        )
        # CASE: From string
        test_model = cobra_core.Model(NAME)
        test_string = ["HOMOMETHIONINE, c"]
        cr.add_metabolites(
            model=test_model,
            obj=test_string,
            directory=dir_data,
            database="META",
        )
        self.assertIn(
            member="HOMOMETHIONINE_c",
            container=[member.id for member in test_model.metabolites],
        )
        # CASE: From List, Custom
        test_model = cobra_core.Model(NAME)
        test_string = ["Custom_c, Custom metabolite, c, H20, 0 "]
        cr.add_metabolites(
            model=test_model,
            obj=test_string,
            directory=dir_data,
            database=None,
        )
        self.assertIn(
            member="Custom_c",
            container=[member.id for member in test_model.metabolites],
        )
        # CASE: multiple entries
        test_model = cobra_core.Model(NAME)
        test_list = ["HOMOMETHIONINE, c", "MALTOSE, c"]
        cr.add_metabolites(
            model=test_model,
            obj=test_list,
            directory=dir_data,
            database="META",
        )
        test_names = ["HOMOMETHIONINE_c", "MALTOSE_c"]
        self.assertListEqual(
            list1=[member.id for member in test_model.metabolites],
            list2=test_names,
        )
        # CASE: Multiple metabolites objects
        test_model = cobra_core.Model(NAME)
        test_list = [
            textbook_kegg.metabolites.get_by_id(item)
            for item in ("C00001_c", "C00002_c", "C00003_c")
        ]
        cr.add_metabolites(model=test_model, obj=test_list)
        for item in ("C00001_c", "C00002_c", "C00003_c"):
            self.assertIn(
                member=item,
                container=[member.id for member in test_model.metabolites],
            )
        # CASE: BIGG database
        test_model = cobra_core.Model(NAME)
        cr.add_metabolites(
            model=test_model,
            obj=["acald, c"],
            directory=dir_data,
            database="BIGG",
            model_id="universal",
        )
        self.assertIn(
            member="acald_c",
            container=[member.id for member in test_model.metabolites],
        )
        # CASE: Using replacement dictionary
        test_model = cobra_core.Model(NAME)
        test_list = ["HOMOMETHIONINE, c", "MALTOSE, c"]
        cr.add_metabolites(
            model=test_model,
            obj=test_list,
            directory=dir_data,
            database="META",
            replacement={"HOMOMETHIONINE": "WATER"},
        )
        self.assertIn(
            member="WATER_c",
            container=[member.id for member in test_model.metabolites],
        )
        self.assertNotIn(
            member="HOMOMETHIONINE_c",
            container=[member.id for member in test_model.metabolites],
        )
        self.assertIn(
            member="MALTOSE_c",
            container=[member.id for member in test_model.metabolites],
        )

    def test_add_reactions(self):
        # CASE: From str
        test_model = cobra_core.Model(NAME)
        cr.add_reactions(
            model=test_model,
            obj=str(dir_input.joinpath("reactions_normal.txt")),
            directory=dir_data,
            database="META",
        )
        for reaction in ("GLC_cb", "RXN_14462_p"):
            self.assertIn(
                member=reaction,
                container=[reaction.id for reaction in test_model.reactions],
            )
        # CASE: From Path
        test_model = cobra_core.Model(NAME)
        cr.add_reactions(
            model=test_model,
            obj=dir_input.joinpath("reactions_normal.txt"),
            directory=dir_data,
            database="META",
        )
        for reaction in ("GLC_cb", "RXN_14462_p"):
            self.assertIn(
                member=reaction,
                container=[reaction.id for reaction in test_model.reactions],
            )
        # CASE: Custom reaction
        test_model = cobra_core.Model(NAME)
        cr.add_reactions(
            model=test_model,
            obj=["GLC_cb, Glucose Transport|GLC_c <-> GLC_b"],
            directory=dir_data,
            database="META",
        )
        self.assertIn(
            member="GLC_cb",
            container=[reaction.id for reaction in test_model.reactions],
        )

        # CASE: From string, custom metabolites
        test_model = cobra_core.Model(NAME)
        cr.add_reactions(
            model=test_model,
            obj=["Custom_cb, Custom reaction|Custom_c <=> Custom_b"],
            directory=dir_data,
            database=None,
        )
        self.assertIn(
            member="Custom_cb",
            container=[reaction.id for reaction in test_model.reactions],
        )
        for metabolite in ["Custom_c", "Custom_b"]:
            self.assertIn(
                member=metabolite,
                container=[meta.id for meta in test_model.metabolites],
            )
        # CASE: From string, KEGG
        test_model = cobra_core.Model(NAME)
        cr.add_reactions(
            model=test_model,
            obj=["R02736, c"],
            directory=dir_data,
            database="KEGG",
            genome="hsa",
        )
        self.assertIn(
            member="R02736_c",
            container=[reaction.id for reaction in test_model.reactions],
        )
        # CASE: From List of strings
        test_model = cobra_core.Model(NAME)
        test_list = [
            "GLC_cb, Glucose Transport|GLC_c <-> GLC_b",
            "RXN-14462, c",
        ]
        cr.add_reactions(
            model=test_model,
            obj=test_list,
            directory=dir_data,
            database="GCF_000020025",
        )
        for reaction in ("GLC_cb", "RXN_14462_c"):
            self.assertIn(
                member=reaction,
                container=[reaction.id for reaction in test_model.reactions],
            )
        self.assertIn(
            member="NPUN_RS12370",
            container=[gene.id for gene in test_model.genes],
        )

        # CASE: multiple reactions
        test_model = cobra_core.Model(NAME)
        test_list = [
            textbook_kegg.reactions.get_by_id(reaction)
            for reaction in ("ACALDt", "ATPS4r", "ACt2r")
        ]
        cr.add_reactions(model=test_model, obj=test_list, directory=dir_data)
        for reaction in ("ACALDt", "ATPS4r", "ACt2r"):
            self.assertIn(
                member=reaction,
                container=[reaction.id for reaction in test_model.reactions],
            )

        # CASE: BIGG
        test_model = cobra_core.Model(NAME)
        cr.add_reactions(
            model=test_model,
            obj=["ACALDt, c"],
            directory=dir_data,
            database="BIGG",
            model_id="universal",
        )
        self.assertIn(
            member="ACALDt_c",
            container=[reaction.id for reaction in test_model.reactions],
        )
        # CASE: Duplicate element
        test_model = textbook_kegg
        cr.add_reactions(
            model=test_model,
            obj=["R08549, c"],
            directory=dir_data,
            database="KEGG",
        )
        cr.add_reactions(
            model=test_model,
            obj=["AKGDH, c"],
            directory=dir_data,
            database="BIGG",
            model_id="universal",
        )
        self.assertNotIn(
            member="AKGDH_c",
            container=[reaction.id for reaction in test_model.reactions],
        )
        # CASE: replacement dicts
        test_model = cobra_core.Model(NAME)
        test_list = [
            "GLC_cb, Glucose Transport|Not_GLC_c <-> GLC_b",
            "RXN-14462, c",
            "ACETALD-DEHYDROG-RXN ,c",
            "AMONITRO-RXN, c",
        ]
        cr.add_reactions(
            model=test_model,
            obj=test_list,
            directory=dir_data,
            database="META",
            replacement={
                "Not_GLC": "GLC",
                "ACETALD-DEHYDROG-RXN": "ACETALD-DEHYDROG-RXN-NEW",
                "Donor-H2": "Donor_Water",
            },
        )
        self.assertTrue(
            test_model.reactions.has_id("GLC_cb"),
        )
        self.assertTrue(
            test_model.reactions.has_id("ACETALD_DEHYDROG_RXN_NEW_c"),
        )
        self.assertFalse(
            test_model.reactions.has_id("Not_GLC_c"),
        )
        for metabolite in ("GLC_b", "GLC_c", "Donor_Water_c"):
            self.assertTrue(
                test_model.metabolites.has_id(metabolite),
                msg=f"{metabolite} not found",
            )

        test_model = cobra_core.Model(NAME)
        test_list = [
            "GLC_cb, Glucose Transport| GLC_c <-> GLC_b",
            "RXN_17742_c, RXN_17742_c |"
            + "1 Oxidized-ferredoxins_c <-> 1 Reduced-ferredoxins_c ",
        ]
        cr.add_reactions(model=test_model, obj=test_list, directory=dir_data)

        for meta in ("GLC_cb", "RXN_17742_c"):
            self.assertTrue(test_model.reactions.has_id(meta))

        # CASE PMN:ARA
        test_model = cobra_core.Model(NAME)
        reaction_str = [
            "PYR_MAL_pc, Pyruvate/Malate transporter |"
            "PYRUVATE_p + MAL_c <-> PYRUVATE_c + MAL_p"
        ]

        cr.add_reactions(test_model, reaction_str, dir_data, "PMN:ARA")

        self.assertTrue(test_model.reactions.has_id("PYR_MAL_pc"))


if __name__ == "__main__":
    print(f"CobraMod version: {cmod_version}")
    print(f"COBRApy version: {cobra_version}")

    unittest.main(verbosity=2, failfast=True)



from collections import UserDict
from contextlib import suppress

from cobramod.error import FoundInPairError


class PairDictionary(UserDict):
    """
    Dictionary that can include a pair (linked) dictionary. Both dictionaries
    cannot have the same keys. It is posible that only a dictionary is a
    PairDictionary, while the other a regular dictionary. The method
    :func:`cobramod.visualization.PairDictionary.set_pair' sets the pair
    for the dictionary.
    """

    def __init__(self, pair: UserDict = None, **kwargs):
        """
        Creates the regular dictionary and can set a pair dictionary if given.
        All keyword-arguments from a regular dictionary can be passed to this
        method.
        """
        self.pair = pair
        super().__init__(**kwargs)

    def __setitem__(self, key, value):
        """
        Set given value for given key in the PairDictionary. It will look
        in the keys of the pair and raise en FoundInPairError.
        """
        # Ignore None
        with suppress(AttributeError):
            if key in self.pair.keys():
                raise FoundInPairError(f"Pair dictionary has {key} as a key")
        super().__setitem__(key=key, item=value)

    def set_pair(self, pair: UserDict):
        """
        Set new pair for the dictionary. If will search for conflicts and raise
        a FoundInPairError if both PairDictionary have common keys.
        """
        # First check for problems and then set pair
        self._find_conflict(pair=pair)
        self.pair = pair

    def _find_conflict(self, pair: UserDict):
        """
        Takes the data of the PairDictionary and its pair and compares it. It
        will raise a FoundInPairError if both dictionaries share common keys.
        """
        first = set(self.data.keys())
        second = set(pair.keys())
        if first.intersection(second):
            raise FoundInPairError



"""Data parsing for BiGG

This module handles the retrieval of data from BiGG into a local directory.
The possible type of data that can be downloaded:

- Metabolites: Normally, simple names.
- Reactions: Mostly abbreviations.
- Genes: It is included in the Reactions. Names and gene-reaction-rule is also
acquired

They change identifiers depending on the model given. BiGG have multiple models
"""
from __future__ import annotations

from contextlib import suppress
from typing import Any

import requests

import cobramod.utils as cmod_utils
from cobramod.debug import debug_log


def find_url(model_id: str, query: str) -> tuple[requests.Response, str]:
    """
    Tries to find a valid URL for the API of BIGG. It will return a
    :class:`requests.Response` if URL is valid.
    object

    Args:
        model_id (str): Name for the specific BIGG model identifier.
        identifier (str): Identifier for the item. Can be a reaction or
            compound

    Returns:
        (Response,str): A valid request Response followed by the database
            Version.

    Raises:
        HTTPError: If identifier is not found in BIGG database.
    """
    for object_type in ("reactions", "metabolites"):
        with suppress(requests.HTTPError):
            # Check that status is available
            debug_log.debug(
                f'{object_type.capitalize()[:-1]} "{query}" not found in '
                f'directory "BIGG", subdirectory "{model_id}".'
            )
            # FIXME: BIGG combines compartments in their internal identifier.
            # To overcome this, use universal
            # if object_type == "metabolites":
            #     model_id = "universal"

            # Retrieve from URL
            url_text = (
                f"http://bigg.ucsd.edu/api/v2/models/{model_id}/{object_type}"
                f"/{query}"
            )
            debug_log.debug(f"Searching {url_text} for biochemical data.")
            # Get and check for errors
            response = requests.get(url_text)
            response.raise_for_status()

            info_response = requests.get(
                "http://bigg.ucsd.edu/api/v2/database_version"
            )
            info_response.raise_for_status()
            db_version = info_response.json()["bigg_models_version"]

            return response, db_version
    # Otherwise
    raise requests.HTTPError(
        f"Identifier '{query}' not found in BIGG (model: '{model_id}')."
    )


def build_reference(json_data: dict[str, Any]) -> dict[str, str]:
    """
    Return a dictionary of cross-references, where the keys are the
    cross-references and values the their identifiers.
    """
    references = json_data["database_links"]
    try:
        return {item: references[item][0]["id"] for item in references}
    except KeyError:
        raise KeyError("Problem with given json dictionary. Inform maintainers")


def parse_reaction_attributes(
    data: dict[str, Any], entry: str
) -> dict[str, Any]:
    results = data.get("results", None)

    if results:
        databases = data.get("database_links", None)
        data = results.pop()
        data["database_links"] = databases

    reaction_str = data.get("reaction_string", "")

    reaction_str = reaction_str.replace("&#8652;", "<->")
    reaction_str = reaction_str.replace("&#x2192;", "-->")
    reaction_str = reaction_str.replace("&rarr;", "-->")
    reaction_str = reaction_str.replace("&larr;", "<--")

    position = cmod_utils.get_arrow_position(reaction_str)

    if position == -1:
        raise AttributeError(
            f"Reaction string for {entry} did not include an arrow in the "
            "equation"
        )

    reactants = [
        item.rstrip().strip()
        for item in reaction_str[:position].rstrip().strip().split("+")
    ]
    products = [
        item.rstrip().strip()
        for item in reaction_str[position + 3 :].rstrip().strip().split("+")
    ]

    arrow = reaction_str[position : position + 3]

    reaction_str = str()

    for side in (reactants, products):
        prefix = "r"

        if side is reactants:
            prefix = "l"

        for i, item in enumerate(side):
            try:
                coefficient, identifier = [
                    item.strip().rstrip() for item in item.split(" ")
                ]
            except ValueError:
                # In case no coefficient is given, it must be 1
                coefficient = "1"
                identifier = item

            reaction_str += f"{coefficient} {prefix}_{identifier[:-2]}"

            if i < len(side) - 1:
                reaction_str += " + "

            if side is reactants and i == len(side) - 1:
                reaction_str += f" {arrow} "

    genes = parse_genes(
        data.get("genes", []), data.get("gene_reaction_rule", "")
    )
    attributes: dict[str, Any] = {
        "name": data.get("name", data.get("exported_reaction_id", "")),
        "equation": reaction_str,
        "xref": build_reference(data),
        "replacements": {},
        "transport": cmod_utils.is_transport(reaction_str),
        "genes": genes,
    }
    return attributes


def parse_metabolite_attributes(data: dict[str, Any]) -> dict[str, Any]:
    # For generic compounds, a list comes, otherwise it is a regular
    # string
    try:
        formula: str = data.get("formulae", data.get("formula", [])).pop()
        charge = int(data["charges"].pop())

    except AttributeError:
        # NOTE: add message
        formula = data.get("formulae", data.get("formula", "X"))
        charge = int(data["charge"])

    attributes = {
        "name": data.get("name", ""),
        "formula": formula,
        "charge": charge,
        "xref": build_reference(data),
    }
    return attributes


def parse_genes(data: list[dict[str, Any]], rule: str):
    genes: dict[str, str] = {}
    for gene in data:
        genes[gene["bigg_id"]] = gene.get("name", "")

    return {"genes": genes, "rule": rule}



#!/usr/bin/env python3
"""Unittest for module pathway

This module includes the TestCase TestGroup. This checks the behavior of the
new child of :obj:`cobra.core.group.Group` "Pathway". This new class is able
to use Escher for its visualizations.
"""
import unittest
from json import loads
from pathlib import Path
from random import randint
from tempfile import TemporaryDirectory
from time import sleep
from webbrowser import open as web_open

import cobra.core as cobra_core
import cobra.io as cobra_io
import cobramod.error as cmod_error
import cobramod.test as cmod_test
import escher
import pandas as pd
from cobra import __version__ as cobra_version
from cobramod import __version__ as cmod_version
from cobramod.core import pathway as pt
from cobramod.core.extension import add_pathway
from cobramod.debug import change_to_debug

change_to_debug()

dir_data = Path(__file__).resolve().parent.joinpath("data")
dir_input = Path(__file__).resolve().parent.joinpath("input")

# If data is missing, then do not test. Data should always be the same
if not dir_data.exists():
    raise NotADirectoryError("Data for the test is missing")


class TestGroup(unittest.TestCase):
    def test___init__(self):
        # CASE: regular __init__
        test_group = pt.Pathway(id="test_group")
        self.assertIn(member="solution", container=dir(test_group))
        self.assertIn(member="members", container=dir(test_group))
        self.assertIn(member="kind", container=dir(test_group))
        self.assertIn(member="id", container=dir(test_group))
        self.assertIn(member="KIND_TYPES", container=dir(test_group))

        # CASE: two entities cannot be the same
        test_group2 = pt.Pathway(id="test_group")
        self.assertIsNot(expr1=test_group, expr2=test_group2)

    def test__filter(self):
        # Configuration
        test_model = cmod_test.textbook_kegg.copy()
        members = {
            test_model.reactions.R00127_c,
            test_model.reactions.R00315_c,
        }
        test_group = pt.Pathway(id="test_group", members=members)
        test_model.add_groups([test_group])

        # CASE: Fluxes
        test_pathway = test_model.groups.get_by_id("test_group")
        if not isinstance(test_pathway, pt.Pathway):
            raise TypeError("Not a valid CobraMod Pathway")

        test_serie = test_pathway._filter(
            solution=test_model.optimize(), attribute="fluxes"
        )
        self.assertEqual(first=test_serie["R00127_c"], second=0)

        # CASE: reduced prices
        test_pathway = test_model.groups.get_by_id("test_group")
        if not isinstance(test_pathway, pt.Pathway):
            raise TypeError("Not a valid CobraMod Pathway")

        test_serie = test_pathway._filter(
            solution=test_model.optimize(), attribute="reduced_costs"
        )
        self.assertEqual(first=test_serie["R00127_c"], second=0)

    def test_solution(self):
        # Configuration
        test_model = cmod_test.textbook_kegg.copy()
        members = {
            test_model.reactions.R00127_c,
            test_model.reactions.R00315_c,
        }

        test_group = pt.Pathway(id="test_group", members=members)
        test_model.add_groups([test_group])
        test_solution = test_group.solution(solution=test_model.optimize())
        # CASE: Regular fluxes

        for attribute in (
            test_solution.fluxes,
            test_solution.reduced_costs,
        ):
            if not isinstance(attribute, pd.Series):
                raise TypeError("Not a valid pandas.Series")

            self.assertTrue(expr="R00127_c" in attribute)
            self.assertTrue(expr="R00315_c" in attribute)

    def test_add_members(self):
        test_model = cmod_test.textbook_kegg.copy()
        # CASE: Regular reactions
        test_group = pt.Pathway(id="test_group")

        test_group.add_members(
            new_members=[
                test_model.reactions.R00127_c,
                test_model.reactions.R00315_c,
            ]
        )
        test_list = [reaction.id for reaction in test_group.members]
        self.assertIn(member="R00127_c", container=test_list)
        self.assertIn(member="R00315_c", container=test_list)
        self.assertDictEqual(
            d1=test_group.notes["ORDER"],
            d2={"R00315_c": None, "R00127_c": None},
        )

        # CASE: Mixing reactions and metabolites
        test_group = pt.Pathway(id="test_group")
        self.assertRaises(
            TypeError,
            test_group.add_members,
            new_members=[
                test_model.reactions.R00127_c,
                test_model.metabolites.C00084_c,
            ],
        )

    def test__transform(self):
        # CASE: Outside the model
        test_model = cmod_test.textbook_kegg.copy()
        members = {
            test_model.reactions.R00127_c,
            test_model.reactions.R00315_c,
        }
        test_group = cobra_core.Group(id="test_group", members=members)
        test_group = pt.Pathway._transform(obj=test_group)
        self.assertIsInstance(obj=test_group, cls=pt.Pathway)

        # CASE: Inside the model
        test_model = cmod_test.textbook_kegg.copy()
        members = {
            test_model.reactions.R00127_c,
            test_model.reactions.R00315_c,
        }
        test_group = cobra_core.Group(id="test_group", members=members)
        test_model.add_groups([test_group])
        test_model.groups[0] = pt.Pathway._transform(obj=test_group)

        self.assertIsInstance(obj=test_model.groups[0], cls=pt.Pathway)

        # CASE: Inside the model, multiple
        test_model = cmod_test.textbook_kegg.copy()

        test_iterable = [
            test_model.reactions[0],
            test_model.reactions[1],
            test_model.reactions[2],
            test_model.reactions[3],
            test_model.reactions[4],
        ]
        for reaction in test_iterable:
            test_group = cobra_core.Group(id=reaction.id)
            test_group.add_members([reaction])
            test_model.add_groups([test_group])

        for index, item in enumerate(test_model.groups):
            test_model.groups[index] = pt.Pathway._transform(item)

        for group in test_model.groups:
            self.assertIsInstance(obj=group, cls=pt.Pathway)

    def test_visualize(self):
        # CASE: Members with initialization.
        test_model = cmod_test.textbook.copy()
        members = {
            test_model.reactions.EX_glc__D_e,
            test_model.reactions.GLCpts,
            test_model.reactions.G6PDH2r,
            test_model.reactions.PGL,
            test_model.reactions.GND,
        }
        test_group = pt.Pathway(id="test_group", members=members)
        test_group.graph = {
            "EX_glc__D_e": "GLCpts",
            "GLCpts": "G6PDH2r",
            "G6PDH2r": "PGL",
            "PGL": "GND",
            "GND": None,
        }
        test_builder = test_group.visualize()

        web_open("pathway.html")
        sleep(1)
        test_group.vertical = True
        test_builder = test_group.visualize()
        if not isinstance(test_builder, escher.Builder):
            raise TypeError("Escher was not loaded")

        self.assertEqual(
            first=len(loads(test_builder.map_json)[1]["reactions"]),  # type: ignore
            second=5,
        )
        # CASE: Members after initialization.
        test_model = cmod_test.textbook.copy()
        test_group = pt.Pathway(id="test_group")
        for identifier in ("EX_glc__D_e", "GLCpts", "G6PDH2r", "PGL", "GND"):
            reaction = test_model.reactions.get_by_id(identifier)
            if not isinstance(reaction, cobra_core.Reaction):
                raise TypeError("Not a valid COBRApy object")

            test_group.add_members(new_members=[reaction])
        self.assertEqual(
            first=len(loads(test_builder.map_json)[1]["reactions"]),  # type: ignore
            second=5,
        )
        # CASE: From copy of model.
        test_model.add_groups(group_list=[test_group])
        test_model_copy = test_model.copy()
        test_group = test_model_copy.groups.get_by_id("test_group")
        self.assertEqual(
            first=len(loads(test_builder.map_json)[1]["reactions"]),  # type: ignore
            second=5,
        )
        # CASE: Regular Biocyc
        test_model = cmod_test.textbook_biocyc.copy()
        add_pathway(
            model=test_model,
            pathway="SALVADEHYPOX-PWY",
            compartment="c",
            directory=dir_data,
            database="ECO",
            ignore_list=[],
            show_imbalance=False,
        )
        # Test fluxes
        test_pathway = test_model.groups.get_by_id("SALVADEHYPOX-PWY")
        if not isinstance(test_pathway, pt.Pathway):
            raise TypeError("Not a valid CobraMod Pathway")

        self.assertEqual(first=len(test_pathway.members), second=5)
        test_solution = {
            reaction.id: randint(-4, 4) for reaction in test_pathway.members
        }
        test_pathway.color_negative = "red"
        test_pathway.color_positive = "green"
        test_pathway.visualize(solution_fluxes=test_solution)
        web_open("pathway.html")
        sleep(1)

        test_pathway.vertical = True
        test_pathway.visualize(solution_fluxes=test_solution)
        web_open("pathway.html")
        sleep(1)

        # CASE: Regular Biocyc
        add_pathway(
            model=test_model,
            pathway="PWY-1187",
            compartment="c",
            directory=dir_data,
            database="ARA",
            ignore_list=[],
            show_imbalance=False,
        )
        # Test fluxes
        test_pathway = test_model.groups.get_by_id("PWY-1187")
        if not isinstance(test_pathway, pt.Pathway):
            raise TypeError("Not a valid CobraMod Pathway")

        self.assertEqual(first=len(test_pathway.members), second=14)
        self.assertTrue(expr=test_pathway.notes["ORDER"])
        test_solution: dict[str, float] = {
            reaction.id: randint(-4, 4) for reaction in test_pathway.members
        }
        test_pathway.color_negative = "purple"
        test_pathway.color_positive = "blue"
        test_pathway.visualize(solution_fluxes=test_solution)
        web_open("pathway.html")
        sleep(1)
        test_pathway.vertical = True
        test_pathway.visualize(solution_fluxes=test_solution)

    def test_model_convert(self):
        # CASE: regular conversion of Groups
        test_model = cmod_test.textbook.copy()
        test_list = [
            test_model.reactions[0],
            test_model.reactions[1],
            test_model.reactions[2],
            test_model.reactions[3],
            test_model.reactions[4],
        ]
        for reaction in test_list:
            test_group = cobra_core.Group(id=reaction.id)
            test_model.add_groups(group_list=[test_group])

            test_group.add_members(new_members=[reaction])

        pt.model_convert(model=test_model)
        for group in test_model.groups:
            self.assertIsInstance(obj=group, cls=pt.Pathway)

        # Test visualize
        test_pathway = test_model.groups[-1]
        if not isinstance(test_pathway, pt.Pathway):
            raise TypeError("Not a valid CobraMod Pathway")
        test_pathway.visualize()
        web_open("pathway.html")
        sleep(1)

        # CASE: Regular Model
        filename = dir_input.joinpath("test_model.sbml")
        test_model = cobra_io.read_sbml_model(str(filename))
        pt.model_convert(test_model)
        for group in test_model.groups:
            self.assertIsInstance(obj=group, cls=pt.Pathway)
        test_pathway = test_model.groups[-1]

        if not isinstance(test_pathway, pt.Pathway):
            raise TypeError("Not a valid CobraMod Pathway")

        test_pathway.visualize()
        web_open("pathway.html")
        sleep(1)

        # CASE: using add_pathway
        test_model = cmod_test.textbook_biocyc.copy()
        sequence = ["PEPDEPHOS-RXN", "PYRUVFORMLY-RXN", "FHLMULTI-RXN"]
        test_graph = {
            "PEPDEPHOS_RXN_c": "PYRUVFORMLY_RXN_c",
            "PYRUVFORMLY_RXN_c": "FHLMULTI_RXN_c",
            "FHLMULTI_RXN_c": None,
        }
        add_pathway(
            model=test_model,
            pathway=sequence,
            directory=dir_data,
            database="ECOLI",
            compartment="c",
            group="test_group",
        )

        with TemporaryDirectory() as tmp_dir:
            filename = Path(tmp_dir).joinpath("test_model.sbml")
            cobra_io.write_sbml_model(
                cobra_model=test_model, filename=str(filename)
            )

            test_model = cobra_io.read_sbml_model(str(filename))

        pt.model_convert(test_model)
        test_pathway = test_model.groups.get_by_id("test_group")

        if not isinstance(test_pathway, pt.Pathway):
            raise TypeError("Not a valid CobraMod Pathway")

        self.assertDictEqual(d1=test_graph, d2=test_pathway.notes["ORDER"])
        self.assertDictEqual(d1=test_graph, d2=test_pathway.graph)
        test_pathway.visualize()
        web_open("pathway.html")
        sleep(1)

        # CASE: Using a Group
        test_model = cmod_test.textbook_biocyc.copy()

        test_group = cobra_core.Group(id="curated_pathway")
        for reaction in ("GLCpts", "G6PDH2r", "PGL", "GND"):
            test_group.add_members([test_model.reactions.get_by_id(reaction)])
        test_model.add_groups([test_group])

        # Conversion to a Pathway
        pt.model_convert(model=test_model)
        test_pathway = test_model.groups.get_by_id("curated_pathway")
        if not isinstance(test_pathway, pt.Pathway):
            raise TypeError("Not a valid CobraMod Pathway")

        test_pathway.vertical = True
        test_pathway.visualize()
        web_open("pathway.html")

    def test_modify_graph(self):
        test_model = cmod_test.textbook_kegg.copy()
        # CASE: merging test
        test_sequence = ["RXN-2206", "RXN-11414", "RXN-11422", "RXN-11430"]
        add_pathway(
            model=test_model,
            pathway=test_sequence,
            database="ARA",
            directory=dir_data,
            compartment="c",
            show_imbalance=False,
            avoid_list=["RXN-2206"],
        )
        test_sequence = ["RXN-11438", "RXN-2208", "RXN-2209", "RXN-2221"]
        add_pathway(
            model=test_model,
            pathway=test_sequence,
            database="ARA",
            directory=dir_data,
            compartment="c",
            show_imbalance=False,
        )
        test_group = test_model.groups.get_by_id("custom_group")

        if not isinstance(test_group, pt.Pathway):
            raise TypeError("Not a valid CobraMod Pathway")

        test_group.modify_graph(
            reaction="RXN_11430_c", next_reaction="RXN_11438_c"
        )
        subset = {"RXN_11430_c": "RXN_11438_c"}
        self.assertEqual(test_group.graph, test_group.graph | subset)
        self.assertEqual(test_group.graph, test_group.notes["ORDER"] | subset)
        test_group.visualize()
        web_open("pathway.html")
        sleep(1)

        # CASE: Keys do not exists
        self.assertRaises(
            cmod_error.GraphKeyError,
            test_group.modify_graph,
            reaction="RXN_11438_c",
            next_reaction="NOT_RXN",
        )
        self.assertRaises(
            cmod_error.GraphKeyError,
            test_group.modify_graph,
            reaction="NOT_RXN",
            next_reaction="RXN_11438_c",
        )
        # CASE: Using None

        subset = {"RXN_11430_c": None}
        test_group.modify_graph(reaction="RXN_11430_c", next_reaction=None)
        self.assertEqual(test_group.graph, test_group.graph | subset)
        self.assertEqual(
            test_group.notes["ORDER"], test_group.notes["ORDER"] | subset
        )


if __name__ == "__main__":
    print(f"CobraMod version: {cmod_version}")
    print(f"COBRApy version: {cobra_version}")

    unittest.main(verbosity=2, failfast=True)



#!/usr/bin/env python3
"""Unit-test for package parsing

This test handles the retrieval of data from a local directory or from
the different databases. The module is separated according to the databases:

- TestBiocyc: For Biocyc-related databases
- TestBigg: For BiGG database
- TestKegg: For Kegg
- TestPlantCyc: For PlantCyc
- TestSolcyc: For SolCyc
"""
import unittest
from pathlib import Path

import cobra.core as cobra_core
import cobramod.error as cmod_error
import cobramod.retrieval as cmod_retrieval
import requests
from cobra import __version__ as cobra_version
from cobramod import __version__ as cmod_version
from cobramod.debug import change_to_debug
from cobramod.parsing import bigg as bi
from cobramod.parsing.db_version import DataVersionConfigurator

dir_data = Path(__file__).resolve().parent.joinpath("data")

change_to_debug()
data_conf = DataVersionConfigurator()

# If data is missing, then do not test. Data should always be the same
if not dir_data.exists():
    raise NotADirectoryError("Data for the test is missing")


class TestKegg(unittest.TestCase):
    """
    Parsing and retrival for Kegg
    """

    @classmethod
    def setUpClass(cls):
        data_conf.force_same_version = True

    def test_get_data(self):
        # CASE: Wrong identifier
        self.assertRaises(
            requests.HTTPError,
            cmod_retrieval.get_data,
            directory=dir_data,
            identifier="R08618-wrong",
            database="KEGG",
        )

        # CASE: Regular reaction
        test_data = cmod_retrieval.get_data(
            directory=dir_data, identifier="R08618", database="KEGG"
        )
        self.assertIsInstance(obj=test_data, cls=cmod_retrieval.Data)

        # CASE: Regular compound
        test_data = cmod_retrieval.get_data(
            directory=dir_data, identifier="C01290", database="KEGG"
        )
        self.assertIsInstance(obj=test_data, cls=cmod_retrieval.Data)

        # CASE: RNA
        test_data = cmod_retrieval.get_data(
            directory=dir_data, identifier="GLT-tRNAs", database="pmn:PLANT"
        )
        self.assertIsInstance(obj=test_data, cls=cmod_retrieval.Data)

        # CASE: Module
        test_data = cmod_retrieval.get_data(
            directory=dir_data, identifier="M00001", database="KEGG"
        )
        test_graph = test_data.attributes["pathway"]
        self.assertEqual(first=test_graph["R00200"], second=None)
        self.assertEqual(first=len(test_graph), second=15)

        # CASE: Simple branch
        test_data = cmod_retrieval.get_data(
            directory=dir_data, identifier="M00040", database="KEGG"
        )
        test_graph = test_data.attributes["pathway"]
        self.assertEqual(first=test_graph["R00732"], second=None)
        self.assertEqual(first=test_graph["R00733"], second=None)
        self.assertEqual(first=len(test_graph), second=4)

    def test_parse(self):
        # CASE: Reaction (same as setup)
        test_data = cmod_retrieval.get_data(
            directory=dir_data,
            identifier="R08618",
            database="KEGG",
            genome="ath",
        )
        self.assertEqual(test_data.attributes["genes"]["rule"], "AT3G19710")
        self.assertEqual(
            test_data.attributes["genes"]["genes"], {"AT3G19710": ""}
        )

        test_reaction = test_data.parse(cobra_core.Model(""), "c")
        self.assertEqual(test_reaction.id, "R08618_c")
        if not isinstance(test_reaction, cobra_core.Reaction):
            raise TypeError("Not a valid COBRApy Reaction")

        self.assertEqual(first=len(test_reaction.metabolites), second=4)
        self.assertEqual(len(test_reaction.genes), 1)

        # CASE: Compound
        test_data = cmod_retrieval.get_data(
            directory=dir_data, identifier="C01290", database="KEGG"
        )
        test_metabolite = test_data.parse(cobra_core.Model(""), "c")
        self.assertEqual(test_metabolite.id, "C01290_c")

        if not isinstance(test_metabolite, cobra_core.Metabolite):
            raise TypeError("Not a valid COBRApy Reaction")
        self.assertEqual(test_metabolite.formula, "C31H56NO13R")

        # CASE: EC number (not working for now)
        self.assertRaises(
            AttributeError,
            cmod_retrieval.get_data,
            directory=dir_data,
            identifier="7.1.2.2",
            database="KEGG",
        )

        # CASE: Pathway
        test_data = cmod_retrieval.get_data(
            directory=dir_data, identifier="M00001", database="KEGG"
        )
        self.assertEqual("Pathway", test_data.mode)
        self.assertEqual("M00001", test_data.identifier)
        self.assertEqual(15, len(test_data.attributes["pathway"]))


class TestBiocyc(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        data_conf.force_same_version = True

    def test_get_data(self):
        # CASE: ID not found
        self.assertRaises(
            requests.HTTPError,
            cmod_retrieval.get_data,
            directory=dir_data,
            identifier="WATER_fake",
            database="META",
        )
        # CASE: Regular META
        test_data = cmod_retrieval.get_data(
            directory=dir_data, identifier="WATER", database="META"
        )
        self.assertIsInstance(obj=test_data, cls=cmod_retrieval.Data)

        test_data = cmod_retrieval.get_data(
            directory=dir_data, identifier="ACETALD", database="ARA"
        )
        self.assertIsInstance(obj=test_data, cls=cmod_retrieval.Data)
        test_data = cmod_retrieval.get_data(
            directory=dir_data, identifier="PWY-5337", database="ARA"
        )
        self.assertIsInstance(obj=test_data, cls=cmod_retrieval.Data)

        # CASE: Regular complex lineal pathway

        test_data = cmod_retrieval.get_data(
            directory=dir_data, identifier="PWY-1187", database="META"
        )

        test_graph = test_data.attributes["pathway"]
        self.assertEqual(len(test_graph), 14)
        self.assertCountEqual(test_graph["RXN-2221"], ("RXN-2222", "RXN-2223"))

        # CASE: Complex cyclic pathway

        test_data = cmod_retrieval.get_data(
            directory=dir_data, identifier="CALVIN-PWY", database="META"
        )
        test_graph = test_data.attributes["pathway"]
        self.assertEqual(len(test_graph), 13)
        self.assertCountEqual(
            test_graph["TRIOSEPISOMERIZATION-RXN"],
            ("SEDOBISALDOL-RXN", "F16ALDOLASE-RXN"),
        )

        # CASE: Single-reaction pathway
        test_data = cmod_retrieval.get_data(
            directory=dir_data, identifier="PWY-7344", database="META"
        )

        test_graph = test_data.attributes["pathway"]
        self.assertEqual(len(test_graph), 1)
        self.assertEqual(test_graph["UDPGLUCEPIM-RXN"], None)

        # CASE: Super-Pathway
        self.assertRaises(
            cmod_error.SuperpathwayException,
            cmod_retrieval.get_data,
            identifier="PWY-5052",
            database="META",
            directory=dir_data,
        )

    def test_parse(self):
        # CASE: Compound
        test_data = cmod_retrieval.get_data(
            directory=dir_data, identifier="AMP", database="META"
        )
        test_metabolite = test_data.parse(cobra_core.Model(""), "p")

        self.assertEqual(test_metabolite.id, "AMP_p")

        if not isinstance(test_metabolite, cobra_core.Metabolite):
            raise TypeError("Not a valid COBRApy Reaction")
        self.assertEqual(test_metabolite.formula, "C10H12N5O7P1")

        # CASE: Reaction in ARA
        test_data = cmod_retrieval.get_data(
            directory=dir_data,
            identifier="GTP-CYCLOHYDRO-II-RXN",
            database="ARA",
        )

        self.assertEqual(
            test_data.attributes["genes"]["rule"], "AT5G64300 or AT5G59750"
        )
        self.assertCountEqual(
            test_data.attributes["genes"]["genes"].keys(),
            ["AT5G64300", "AT5G59750"],
        )

        # NOTE: retrieve data before this test
        test_reaction = test_data.parse(cobra_core.Model(""), "c")
        self.assertEqual(test_reaction.id, "GTP_CYCLOHYDRO_II_RXN_c")

        if not isinstance(test_reaction, cobra_core.Reaction):
            raise TypeError("Not a valid COBRApy Reaction")
        self.assertEqual(len(test_reaction.genes), 2)

        # CASE: Reaction in Meta
        test_data = cmod_retrieval.get_data(
            directory=dir_data,
            identifier="GTP-CYCLOHYDRO-II-RXN",
            database="META",
        )
        test_reaction = test_data.parse(cobra_core.Model(""), "c")
        self.assertEqual(test_reaction.id, "GTP_CYCLOHYDRO_II_RXN_c")
        self.assertEqual(test_data.attributes["genes"]["rule"], "")
        self.assertCountEqual(test_data.attributes["genes"]["genes"].keys(), [])
        if not isinstance(test_reaction, cobra_core.Reaction):
            raise TypeError("Not a valid COBRApy Reaction")
        self.assertEqual(len(test_reaction.genes), 0)

        self.assertEqual(test_reaction.bounds, (0, 1000))

        # CASE: Reaction in small subdatabase
        test_data = cmod_retrieval.get_data(
            directory=dir_data,
            identifier="GTP-CYCLOHYDRO-II-RXN",
            database="GCF_000010885",
        )
        self.assertCountEqual(
            test_data.attributes["genes"]["genes"].keys(),
            ["APA22_RS09400"],
        )
        test_reaction = test_data.parse(cobra_core.Model(""), "c")
        if not isinstance(test_reaction, cobra_core.Reaction):
            raise TypeError("Not a valid COBRApy Reaction")
        self.assertEqual(test_reaction.bounds, (0, 1000))

        # CASE: Protein
        test_data = cmod_retrieval.get_data(
            directory=dir_data,
            identifier="Reduced-hemoproteins",
            database="ARA",
        )

        test_metabolite = test_data.parse(cobra_core.Model(""), "p")

        self.assertEqual(test_metabolite.id, "Reduced_hemoproteins_p")

        if not isinstance(test_metabolite, cobra_core.Metabolite):
            raise TypeError("Not a valid COBRApy Reaction")

        self.assertEqual(test_metabolite.formula, "X")

        # CASE: Pathway
        test_data = cmod_retrieval.get_data(
            directory=dir_data,
            identifier="PWY-1187",
            database="META",
        )
        test_graph = test_data.attributes["pathway"]
        self.assertEqual(len(test_graph), 14)


class TestPlantCyc(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        data_conf.force_same_version = True

    def test_get_data(self):
        # CASE: ID not found
        self.assertRaises(
            requests.HTTPError,
            cmod_retrieval.get_data,
            directory=dir_data,
            identifier="WATER_fake",
            database="pmn:PLANT",
        )

        # CASE: Regular PLANT
        test_data = cmod_retrieval.get_data(
            directory=dir_data, identifier="WATER", database="pmn:PLANT"
        )
        self.assertIsInstance(obj=test_data, cls=cmod_retrieval.Data)

        # CASE: META from pmn
        test_data = cmod_retrieval.get_data(
            directory=dir_data, identifier="CPD-15323", database="pmn:META"
        )
        self.assertIsInstance(obj=test_data, cls=cmod_retrieval.Data)

        # CASE: Pathway from CORN
        test_data = cmod_retrieval.get_data(
            directory=dir_data, identifier="CALVIN-PWY", database="pmn:CORN"
        )
        self.assertIsInstance(obj=test_data, cls=cmod_retrieval.Data)

        # CASE: Regular complex lineal pathway
        test_data = cmod_retrieval.get_data(
            directory=dir_data, identifier="PWY-1187", database="pmn:META"
        )
        test_graph = test_data.attributes["pathway"]
        self.assertEqual(len(test_graph), 14)
        self.assertCountEqual(test_graph["RXN-2221"], ("RXN-2222", "RXN-2223"))

        # CASE: Complex cyclic pathway
        test_data = cmod_retrieval.get_data(
            directory=dir_data, identifier="CALVIN-PWY", database="pmn:CORN"
        )
        test_graph = test_data.attributes["pathway"]
        self.assertEqual(len(test_graph), 13)
        self.assertCountEqual(
            test_graph["TRIOSEPISOMERIZATION-RXN"],
            ("SEDOBISALDOL-RXN", "F16ALDOLASE-RXN"),
        )

        # CASE: Single-reaction pathway
        test_data = cmod_retrieval.get_data(
            directory=dir_data, identifier="PWY-7344", database="pmn:META"
        )

        test_graph = test_data.attributes["pathway"]
        self.assertEqual(len(test_graph), 1)
        self.assertEqual(test_graph["UDPGLUCEPIM-RXN"], None)

        # CASE: Super-Pathway
        self.assertRaises(
            cmod_error.SuperpathwayException,
            cmod_retrieval.get_data,
            identifier="PWY-5052",
            database="pmn:META",
            directory=dir_data,
        )

    def test_parse(self):
        # CASE: Compound
        test_data = cmod_retrieval.get_data(
            directory=dir_data, identifier="AMP", database="pmn:META"
        )
        test_metabolite = test_data.parse(cobra_core.Model(""), "p")

        self.assertEqual(test_metabolite.id, "AMP_p")

        if not isinstance(test_metabolite, cobra_core.Metabolite):
            raise TypeError("Not a valid COBRApy Reaction")
        self.assertEqual(test_metabolite.formula, "C10H12N5O7P1")

        # CASE: Reaction in ARA
        test_data = cmod_retrieval.get_data(
            directory=dir_data,
            identifier="GTP-CYCLOHYDRO-II-RXN",
            database="pmn:ARA",
        )

        self.assertEqual(
            test_data.attributes["genes"]["rule"], "AT5G64300 or AT5G59750"
        )
        self.assertCountEqual(
            test_data.attributes["genes"]["genes"].keys(),
            ["AT5G64300", "AT5G59750"],
        )
        test_reaction = test_data.parse(cobra_core.Model(""), "c")
        self.assertEqual(test_reaction.id, "GTP_CYCLOHYDRO_II_RXN_c")

        if not isinstance(test_reaction, cobra_core.Reaction):
            raise TypeError("Not a valid COBRApy Reaction")
        self.assertEqual(len(test_reaction.genes), 2)

        # CASE: Reaction in Meta
        test_data = cmod_retrieval.get_data(
            directory=dir_data,
            identifier="GTP-CYCLOHYDRO-II-RXN",
            database="pmn:META",
        )
        test_reaction = test_data.parse(cobra_core.Model(""), "c")
        self.assertEqual(test_reaction.id, "GTP_CYCLOHYDRO_II_RXN_c")
        self.assertEqual(test_data.attributes["genes"]["rule"], "")
        self.assertCountEqual(test_data.attributes["genes"]["genes"].keys(), [])
        if not isinstance(test_reaction, cobra_core.Reaction):
            raise TypeError("Not a valid COBRApy Reaction")
        self.assertEqual(len(test_reaction.genes), 0)

        self.assertEqual(test_reaction.bounds, (0, 1000))

        # CASE: Reaction in another database
        test_data = cmod_retrieval.get_data(
            directory=dir_data,
            identifier="GTP-CYCLOHYDRO-II-RXN",
            database="pmn:SOY",
        )
        test_reaction = test_data.parse(cobra_core.Model(""), "c")
        if not isinstance(test_reaction, cobra_core.Reaction):
            raise TypeError("Not a valid COBRApy Reaction")
        self.assertEqual(len(test_reaction.genes), 8)

        # CASE: Protein
        test_data = cmod_retrieval.get_data(
            directory=dir_data,
            identifier="Reduced-hemoproteins",
            database="pmn:ARA",
        )
        test_metabolite = test_data.parse(cobra_core.Model(""), "p")

        self.assertEqual(test_metabolite.id, "Reduced_hemoproteins_p")

        if not isinstance(test_metabolite, cobra_core.Metabolite):
            raise TypeError("Not a valid COBRApy Reaction")

        self.assertEqual(test_metabolite.formula, "X")

        # CASE: Pathway
        test_data = cmod_retrieval.get_data(
            directory=dir_data,
            identifier="PWY-1187",
            database="pmn:META",
        )
        test_graph = test_data.attributes["pathway"]
        self.assertEqual(len(test_graph), 14)


class TestBigg(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        data_conf.force_same_version = True

    def test_find_url(self):
        # CASE: Error
        self.assertRaises(
            requests.HTTPError,
            bi.find_url,
            model_id="e_coli_core",
            query="Invalid",
        )

        # CASE: from specific model
        response, _ = bi.find_url(query="ACALD", model_id="e_coli_core")
        self.assertIsInstance(response, requests.Response)

    def test_get_data(self):
        # CASE: metabolite
        test_data = cmod_retrieval.get_data(
            directory=dir_data,
            identifier="accoa_c",
            model_id="e_coli_core",
            database="BIGG",
        )
        self.assertEqual(test_data.mode, "Metabolite")
        self.assertIsInstance(obj=test_data, cls=cmod_retrieval.Data)

        # CASE: Regular reaction from ecoli
        test_data = cmod_retrieval.get_data(
            directory=dir_data,
            identifier="ACALD",
            model_id="e_coli_core",
            database="BIGG",
        )
        self.assertEqual(test_data.mode, "Reaction")
        self.assertIsInstance(obj=test_data, cls=cmod_retrieval.Data)

    def test_parse(self):
        # CASE: Regular universal reaction
        test_data = cmod_retrieval.get_data(
            directory=dir_data,
            identifier="ACALDt",
            model_id="universal",
            database="BIGG",
        )
        test_reaction = test_data.parse(cobra_core.Model(""), "c")

        if not isinstance(test_reaction, cobra_core.Reaction):
            raise TypeError("Not a valid COBRApy Reaction")
        self.assertEqual(len(test_reaction.genes), 0)

        self.assertEqual(first=len(test_reaction.metabolites), second=2)

        # CASE: Regular universal metabolite
        test_data = cmod_retrieval.get_data(
            directory=dir_data,
            identifier="coa",
            model_id="universal",
            database="BIGG",
        )
        test_metabolite = test_data.parse(cobra_core.Model(""), "p")
        if not isinstance(test_metabolite, cobra_core.Metabolite):
            raise TypeError("Not a valid COBRApy Reaction")

        self.assertEqual(test_metabolite.formula, "C21H32N7O16P3S")

        # CASE: Ecoli reaction
        test_data = cmod_retrieval.get_data(
            directory=dir_data,
            identifier="PDH",
            model_id="e_coli_core",
            database="BIGG",
        )
        self.assertEqual(
            test_data.attributes["genes"]["rule"], "b0114 and b0115 and b0116"
        )

        test_reaction = test_data.parse(cobra_core.Model(""), "p")

        if not isinstance(test_reaction, cobra_core.Reaction):
            raise TypeError("Not a valid COBRApy Reaction")

        self.assertEqual(len(test_reaction.genes), 3)

        # CASE: Ecoli metabolite
        test_data = cmod_retrieval.get_data(
            directory=dir_data,
            identifier="co2_c",
            model_id="e_coli_core",
            database="BIGG",
        )
        test_metabolite = test_data.parse(cobra_core.Model(""), "p")
        if not isinstance(test_metabolite, cobra_core.Metabolite):
            raise TypeError("Not a valid COBRApy Reaction")

        self.assertEqual(test_metabolite.id, "co2_p")
        self.assertEqual(test_metabolite.formula, "CO2")


if __name__ == "__main__":
    print(f"CobraMod version: {cmod_version}")
    print(f"COBRApy version: {cobra_version}")

    unittest.main(verbosity=2, failfast=True)



"""Modules genes

This module is responsible for the creation of COBRApy Genes from the parsed
data of the specific parsers.
"""
from typing import Any

import cobra.core as cobra_core

from cobramod.debug import debug_log


def genes_to_reaction(reaction: cobra_core.Reaction, data_dict: dict[str, Any]):
    """
    Adds the corresponding gene-reaction rule to the Reaction and correct
    its name according to the passed dictionary
    """
    rule: str = data_dict["rule"]

    # This line will create the corresponding genes of the reaction
    reaction.gene_reaction_rule = rule

    # Update name
    for gene in reaction.genes:
        gene.name = data_dict["genes"][gene.id]

        if not gene.name:
            gene.name = gene.id

        debug_log.debug(
            f'Gene "{gene.id}" was added with gene name "{gene.name}" to '
            f'reaction "{reaction.id}".'
        )



#!/usr/bin/env python3
"""Unit-test for model behavior

This module checks if a large and a small model behaves as intended. This
examples should simulate real cases. There are two test:

- ShortModel: This should utilize the textbook_biocyc model from cobramod.
- LargeModel: Uses a real GEM
"""
import unittest
from pathlib import Path

import cobra.core as cobra_core
import cobra.io as cobra_io
import cobramod.error as cmod_error
from cobra import __version__ as cobra_version
from cobramod import __version__ as cmod_version
from cobramod.core import extension as ex
from cobramod.core.creation import add_reactions
from cobramod.debug import change_to_debug
from cobramod.parsing.db_version import DataVersionConfigurator
from cobramod.test import textbook_biocyc

change_to_debug()
dir_data = Path(__file__).resolve().parent.joinpath("data")

data_conf = DataVersionConfigurator()

# If data is missing, then do not test. Data should always be the same
if not dir_data.exists():
    raise NotADirectoryError("Data for the test is missing")

dir_input = Path(__file__).resolve().parent.joinpath("input")

# Large Model
main_model = cobra_io.read_sbml_model(
    str(dir_input.joinpath("test_model.sbml"))
)


class ShortModel(unittest.TestCase):
    def test_lineal_pathways(self):
        # For this test, Gluconeogenesis; L-aspartate and L-asparagine
        # biosynthesis, and Nicotine biosynthesis added to test for feasible
        # results.
        test_model = textbook_biocyc.copy()

        # Adding Gluconeogenesis
        ex.add_pathway(
            model=test_model,
            pathway="GLUCONEO-PWY",
            directory=dir_data,
            database="META",
            compartment="c",
            show_imbalance=False,
        )
        self.assertGreater(a=abs(test_model.optimize().objective_value), b=0)
        self.assertIn(
            member="GLUCONEO-PWY",
            container=[group.id for group in test_model.groups],
        )
        add_reactions(
            model=test_model,
            obj=[
                "Redox_Hemoprotein_c, Redox_Hemoprotein_c | "
                + "Ox-NADPH-Hemoprotein-Reductases_c <-> "
                + "Red-NADPH-Hemoprotein-Reductases_c"
            ],
            directory=dir_data,
            database="ARA",
        )

        # Adding Nicotine pathway
        ex.add_pathway(
            model=test_model,
            pathway="PWY-5316",
            directory=dir_data,
            database="META",
            compartment="c",
            ignore_list=[],
            show_imbalance=False,
        )
        self.assertIn(
            member="PWY-5316",
            container=[group.id for group in test_model.groups],
        )

        # Adding aspartate and asparagine biosynthesis
        # Before 27.1 in Metacyc, this was a pathway
        self.assertRaises(
            cmod_error.SuperpathwayException,
            ex.add_pathway,
            model=test_model,
            pathway="ASPASN-PWY",
            directory=dir_data,
            database="META",
            compartment="c",
            ignore_list=[],
            show_imbalance=False,
        )
        ex.add_pathway(
            model=test_model,
            pathway="ASPARAGINESYN-PWY",
            directory=dir_data,
            database="META",
            compartment="c",
            ignore_list=[],
            show_imbalance=False,
            group="ASPASN-PWY",
        )
        ex.add_pathway(
            model=test_model,
            pathway="ASPARTATESYN-PWY",
            directory=dir_data,
            database="META",
            compartment="c",
            ignore_list=[],
            show_imbalance=False,
            group="ASPASN-PWY",
        )
        self.assertGreater(a=abs(test_model.optimize().objective_value), b=0)
        self.assertIn(
            member="ASPASN-PWY",
            container=[group.id for group in test_model.groups],
        )

    def test_cyclical_pathways(self):
        test_model = textbook_biocyc.copy()

        # Adding Mannitol cycle
        ex.add_pathway(
            model=test_model,
            pathway="PWY-6531",
            directory=dir_data,
            database="META",
            compartment="c",
            ignore_list=[],
            show_imbalance=False,
        )
        self.assertIn(
            member="PWY-6531",
            container=[group.id for group in test_model.groups],
        )

        # Adding glyoxylate cycle
        ex.add_pathway(
            model=test_model,
            pathway="GLYOXYLATE-BYPASS",
            directory=dir_data,
            database="ARA",
            compartment="c",
            ignore_list=[],
            show_imbalance=False,
        )
        self.assertGreater(a=abs(test_model.optimize().objective_value), b=0)
        self.assertIn(
            member="GLYOXYLATE-BYPASS",
            container=[group.id for group in test_model.groups],
        )

    def test_multi_compartment(self):
        test_model = textbook_biocyc.copy()
        # This files has transports between cytosol and plastids
        add_reactions(
            model=test_model,
            obj=dir_input.joinpath("test_multi_reactions.txt"),
            database="ARA",
            directory=dir_data,
        )

        ex.add_pathway(
            model=test_model,
            pathway="GLUTATHIONESYN-PWY",
            directory=dir_data,
            database="ARA",
            ignore_list=[],
            compartment="p",
            show_imbalance=False,
        )
        reaction = test_model.reactions.get_by_id("Biomass_Ecoli_core")
        if not isinstance(reaction, cobra_core.Reaction):
            raise TypeError("Not a valid COBRApy Reaction")

        metabolite = test_model.metabolites.get_by_id("GLUTATHIONE_p")
        if not isinstance(metabolite, cobra_core.Metabolite):
            raise TypeError("Not a valid COBRApy metabolite")

        reaction.add_metabolites({metabolite: -1})

        self.assertIn(
            member="GLUTATHIONESYN-PWY",
            container=[group.id for group in test_model.groups],
        )

        self.assertGreater(abs(test_model.optimize().objective_value), 0)

        add_reactions(
            model=test_model,
            obj=["TRANS_GLY_cp, Transport GLY_cp | GLY_c <-> GLY_p"],
            directory=dir_data,
            database="ARA",
        )

        ex.add_pathway(
            model=test_model,
            pathway="PWY-1187",
            directory=dir_data,
            database="ARA",
            ignore_list=["PYRUVATE_c", "CO_A_c", "PROTON_c", "CPD_3746_c"],
            compartment="c",
            show_imbalance=False,
        )
        reaction = test_model.reactions.get_by_id("Biomass_Ecoli_core")
        if not isinstance(reaction, cobra_core.Reaction):
            raise TypeError("Not a valid COBRApy Reaction")

        metabolite_a = test_model.metabolites.get_by_id(
            "3_METHYLSULFINYLPROPYL_GLUCOSINOLATE_c"
        )
        if not isinstance(metabolite_a, cobra_core.Metabolite):
            raise TypeError("Not a valid COBRApy metabolite")
        metabolite_b = test_model.metabolites.get_by_id("GLUTATHIONE_p")
        if not isinstance(metabolite_b, cobra_core.Metabolite):
            raise TypeError("Not a valid COBRApy metabolite")

        reaction.add_metabolites({metabolite_a: -0.75, metabolite_b: 1})
        self.assertGreater(abs(test_model.optimize().objective_value), 0)

        # Lipid initalization
        ex.add_pathway(
            model=test_model,
            pathway="PWY-4381",
            directory=dir_data,
            database="ARA",
            ignore_list=["CO_A_p"],
            compartment="p",
            show_imbalance=False,
        )
        reaction = test_model.reactions.get_by_id("Biomass_Ecoli_core")
        if not isinstance(reaction, cobra_core.Reaction):
            raise TypeError("Not a valid COBRApy Reaction")

        metabolite = test_model.metabolites.get_by_id("Acetoacetyl_ACPs_p")
        if not isinstance(metabolite, cobra_core.Metabolite):
            raise TypeError("Not a valid COBRApy metabolite")

        reaction.add_metabolites({metabolite: -1})

        self.assertGreater(test_model.slim_optimize(error_value=0), 0)


class LargeModel(unittest.TestCase):
    def test_lineal_pathways(self):
        """In this test, unrelated pathways are appended to the large model
        and tested
        """
        test_model = main_model.copy()

        # Autotroph environment
        test_model.reactions.Sucrose_tx.bounds = (0, 0)  # SUCROSE
        test_model.reactions.GLC_tx.bounds = (0, 0)  # GLUCOSE

        # # Photon uptake
        test_model.reactions.Photon_tx.bounds = (-1000, 250)

        # No ammonium
        test_model.reactions.NH4_tx.bounds = (0, 0)
        test_model.objective = "Biomass_tx"

        # Adding methionine metabolism
        ex.add_pathway(
            model=test_model,
            pathway="PWY-7614",
            directory=dir_data,
            database="META",
            compartment="p",
            show_imbalance=False,
        )

        # Optimizing methiin
        test_model.objective = "RXN_8908_p"

        self.assertGreater(test_model.optimize().fluxes["RXN_16201_p"], 0)
        self.assertGreater(test_model.slim_optimize(error_value=0), 0)

        # Adding stachyose biosynthesis
        test_model.objective = "Biomass_tx"

        ex.add_pathway(
            model=test_model,
            pathway="PWY-5337",
            directory=dir_data,
            database="ARA",
            compartment="c",
            show_imbalance=False,
        )

        # Optimizing raffinose
        test_model.objective = "2.4.1.82_RXN_c"

        self.assertGreater(test_model.optimize().fluxes["2.4.1.123_RXN_c"], 0)
        self.assertGreater(test_model.slim_optimize(error_value=0), 0)

        # Adding abscisic acid
        test_model.objective = "Biomass_tx"

        ex.add_pathway(
            model=test_model,
            pathway="PWY-695",
            directory=dir_data,
            database="ARA",
            compartment="p",
            show_imbalance=False,
        )

        # Optimizing abscisate
        test_model.objective = "1.2.3.14_RXN_p"

        self.assertGreater(test_model.optimize().fluxes["1.1.1.288_RXN_p"], 0)
        self.assertGreater(test_model.slim_optimize(error_value=0), 0)

        test_model.objective = "Biomass_tx"
        self.assertGreater(test_model.slim_optimize(error_value=0), 0)


if __name__ == "__main__":
    print(f"CobraMod version: {cmod_version}")
    print(f"COBRApy version: {cobra_version}")

    unittest.main(verbosity=2, failfast=True)
